#include <stdlib.h>
#include <math.h>
#include <stdio.h>

#include "speakerid.h"
#include "matrix.h"
#include "pattern.h"
#include "clustering.h"
#include "distortion.h"

CLUSTERING NewClustering( int CodebookSize, 
			  PATTERN TrainingData ) {

   CLUSTERING CL;
    
   CL = malloc(sizeof(CLUSTERING_TYPE));
   CL->Codebook = NewPattern( PDimensionality(TrainingData), 
			      CodebookSize );
   CL->P = malloc(PNumVectors(TrainingData) * sizeof(int));
   CL->ClusterSize = (int *) malloc(CodebookSize * sizeof(int));
   CL->NumTrainingVectors = PNumVectors(TrainingData);

   return CL;
}

void DeleteClustering(CLUSTERING CL) {

   DeletePattern(CL->Codebook);
   free(CL->P);
   free(CL->ClusterSize);
}

void CopyClustering(CLUSTERING To, CLUSTERING From) {

   int i;

   CopyVectors(To->Codebook, From->Codebook, 1, BookSize(From));

   for (i=1; i <= BookSize(From); i++)
       To->ClusterSize[i-1] = From->ClusterSize[i-1];
 
   for (i=1; i <= NumTrainingVectors(From); i++)
       To->P[i-1] = From->P[i-1];
}

void PrintCodebook(FILE* f, CLUSTERING CL) {

   PrintPattern(GetCodebook(CL), f);
}

void PrintPartition(FILE* f, CLUSTERING CL) {
   
   int i;

   for(i=1; i <= NumTrainingVectors(CL); i++)
       fprintf(f, "%i -> %i\n", i, CL->P[i-1]); 
}

void InitClustering(CLUSTERING CL, PATTERN TrainingData) {

   PickRandomVectors(CL->Codebook, TrainingData);

}

double EvaluateClustering(CLUSTERING CL, PATTERN TrainingData) {

    int i; 
    double Value;
    MATRIX x, c;
    
    x = NewMatrix(PDimensionality(TrainingData), 1);
    c = NewMatrix(PDimensionality(TrainingData), 1);

    Value = 0;	
    for(i=1; i <= PNumVectors(TrainingData); i++) {
       GetVector(x, TrainingData, i);
       GetVector(c, CL->Codebook, CL->P[i-1]);
       Value += VectorDistortion(x, c);
    }

    DeleteMatrix(x); DeleteMatrix(c);
	
    return Value;
}

void MakePartition(CLUSTERING CL, PATTERN TrainingData) {
    
    int i;
    MATRIX x;

    x = NewMatrix(PDimensionality(TrainingData), 1);
    for(i=1; i <= PNumVectors(TrainingData); i++) {
       CL->P[i-1] = FindNearestVector(GetVector(x, TrainingData, i),
			CL->Codebook );   
    }
    DeleteMatrix(x);
}

void ComputeNewCentroids(CLUSTERING CL, PATTERN TrainingData) {

    int i, j, Count;

    MATRIX x, c;

    x = NewMatrix(PDimensionality(TrainingData), 1);
    c = NewMatrix(PDimensionality(TrainingData), 1);

    for(i=1; i <= BookSize(CL); i++) {
       ZeroMatrix(c);
       Count = 0;
       for(j=1; j <= PNumVectors(TrainingData); j++) {
           if (i == CL->P[j-1]) {
	      GetVector(x, TrainingData, j);
	      MatrixSum(c, c, x);
	      Count++;
           }
       }
       if ( Count > 0 ) {
       	   MultiplyByScalar(c, c, 1.0/Count);
	   SetVector(CL->Codebook, c, i);
       }
       CL->ClusterSize[i-1] = Count;
    }	

    DeleteMatrix(x); DeleteMatrix(c);

}

void IterateGLA(CLUSTERING CL, PATTERN TrainingData, Parameters *parap) {
   
    int i, MaxIter; 
    double MinChange, Delta, PrevErr, CurrErr;

    MaxIter = parap->modelptr->max_iterations;
    MinChange = parap->modelptr->stop_threshold;

    MakePartition(CL, TrainingData);
    CurrErr = EvaluateClustering(CL, TrainingData);

    for(i=1; i <= MaxIter; i++) {
        PrevErr = CurrErr;
	MakePartition(CL, TrainingData);
	ComputeNewCentroids(CL, TrainingData);
	CurrErr = EvaluateClustering(CL, TrainingData);
	Delta = PrevErr/CurrErr - 1;
        if (Delta <= MinChange)
	   break;
    }

}

void RGLA(CLUSTERING CL, PATTERN TrainingData, Parameters *parap) {

    int i, NumberOfInitialSolutions;
    double Best, Current;
    CLUSTERING CurrentSolution, BestSolution;


    NumberOfInitialSolutions = parap->modelptr->num_initial_solutions;
    BestSolution = NewClustering(BookSize(CL), TrainingData);
    CurrentSolution = NewClustering(BookSize(CL), TrainingData);
    InitClustering(BestSolution, TrainingData);

    IterateGLA(BestSolution, TrainingData, parap);
    Best = -EvaluateClustering(BestSolution, TrainingData);

    for(i=2; i <= NumberOfInitialSolutions; i++) {

	InitClustering(CurrentSolution, TrainingData);
	IterateGLA(CurrentSolution, TrainingData, parap);
	Current = -EvaluateClustering(CurrentSolution, TrainingData);
	if (Current > Best) {
	   CopyClustering(BestSolution, CurrentSolution);
	   Best = Current;
        }
    }
	
    CopyClustering(CL, BestSolution);
    DeleteClustering(CurrentSolution);
    DeleteClustering(BestSolution);    

}

/* By Ville Hautamki: Tue May 28 16:35:58 EET DST 2002
 * Creates new Pattern and copies all vectors from Clustering to it, 
 * then frees Clustering.
 */

PATTERN Clustering2Pattern(CLUSTERING CL)
{
  PATTERN new;

  new = NewPattern(BookDim(CL), BookSize(CL));
  CopyVectors(new, GetCodebook(CL), 1, BookSize(CL));
  DeleteClustering(CL);

  return new;
}
