/* dct.c by Ville Hautamki <villeh@cs.joensuu.fi>, 2002
 *
 */

#include <float.h>
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "speakerid.h"
#include "util.h"
#include "pattern.h"
#include "matrix.h"
#include "dct.h"

/* --------------------------------------------------------------- */
/*	
	Functioning:Computes discrete cosine transform of log filtered 
                    spectrum.
	Inputs:	 pointer to log filtered spectrum table and pointer to
                 parameter struct. 
						
	Outputs: Returns an transformed vector in MATRIX data structure. 

	Return value:	- 
                                Ville Hautamki, 2002
	History:		Teemu Kilpelinen 26.04.2001
	Version:		1.1
	Reference:		
*/



MATRIX DCT(double *log_filtered_spectrum, Parameters *parap)
{
  int n, k, MFCC, num_mel_filters; 
  double sum;
  MATRIX vector;

  MFCC = parap->featptr->MFCC_ord;
  num_mel_filters = parap->featptr->num_mel_filters;

  vector = NewMatrix(MFCC + 1, 1);

  for(n = 0; n <= MFCC; n++) { 
    sum = 0.0;
    for (k = 1; k <= num_mel_filters; k++) {
      sum = sum + (double)(log_filtered_spectrum[k - 1] * 
			  cos(n * (k - 0.5) * (PI / num_mel_filters)));
    }
    SetMatrixElement(vector, n + 1, 1, sum);
  }

  return vector;
}

/* Removes silent frames from Cepstrum. 
 * Takes as parameters Cepstrum in PATTERN data structure and pointer 
 *  to Parameter struct.
 * Returns new Cepstrum also in PATTERN data structure. Old Cepstrum 
 *  has to be freed on higher level. 
 */

PATTERN remove_silent_frames(PATTERN fromdct, Parameters *parap)
{
  int i, j = 1, include_0th_coeff, remove_sil, MFCC, drop = 0;
  double threshold;
  PATTERN cepstrum;
  MATRIX vector;

  include_0th_coeff = parap->featptr->include_0th_coeff;
  remove_sil = parap->featptr->remove_sil;
  MFCC = parap->featptr->MFCC_ord;
  threshold = parap->featptr->sil_thr * mean_0th(fromdct);

  if ((remove_sil == 1) && (include_0th_coeff == 0)) {
    cepstrum = NewPattern(MFCC, above_threshold(fromdct, threshold));
    printf("With Silence removal and W/O 0th coeff\n");    
    printf("threshold %f\n", threshold);
    printf("above thr %d\n",above_threshold(fromdct, threshold));
 
    for (i = 1; i < PNumVectors(fromdct); i++) {
      if (get_0th(fromdct, i)> threshold) {
	vector = NewMatrix(MFCC, 1);
	copy_vector(fromdct, vector, i, 2);
	SetVector(cepstrum, vector, j);
	j++;
      } else 
	drop++;
    }   

  } else if ((remove_sil == 1) && (include_0th_coeff == 1)) {
    cepstrum = NewPattern(MFCC + 1, above_threshold(fromdct, threshold));
    printf("With Silence removal and With 0th coeff\n");
    printf("threshold %f\n", threshold);
    printf("above thr %d\n",above_threshold(fromdct, threshold));

    for (i = 1; i < PNumVectors(fromdct); i++) {
      if (get_0th(fromdct, i)> threshold) {
	vector = NewMatrix(MFCC + 1, 1);
	copy_vector(fromdct, vector, i, 1);
	SetVector(cepstrum, vector, j);
	j++;
      } else 
	drop++;
    }   

  } else if ((remove_sil == 0) && (include_0th_coeff == 0)) {
    cepstrum = NewPattern(MFCC, PNumVectors(fromdct));

    printf("W/O Silence removal and W/O 0th coeff\n");

    for (i = 1; i < PNumVectors(fromdct); i++) {
      vector = NewMatrix(MFCC, 1);
      copy_vector(fromdct, vector, i, 2);
      SetVector(cepstrum, vector, i);
    }    

  } else { 
    printf("W/O Silence removal and With 0th coeff\n");

    return fromdct;
  }

  if (remove_sil)
    printf("Dropped frames: %d\n", drop);

  return cepstrum;
}

/* Copies one vector component by component from PATTERN to MATRIX data 
 *  structure.
 * Takes as parameters from PATTERN, to MATRIX, position column in PATTERN 
 *  and starting position in copied vector (we either start from first 
 *  scalar or second.
 * Function does not return anything. 
 */

void copy_vector(PATTERN from, MATRIX to, int pos, int startpos)
{
  int i, position;
  double scalar;

  if (startpos == 1) 
    position = 0; 
  else 
    position = 1;

  for (i = 1; i < MRows(to); i++) {
    scalar = GetMatrixElement(from->Data, position + i, pos);
    SetMatrixElement(to, i, 1, scalar);
  }
}

/* Returns 0th coefficient from one vector in cepstrum in PATTERN
 * Takes as parameters Cepstrum in PATTERN data structure, number of 
 *  vector in cepstrum, which 0th coeff is returned.
 * Returns 0th coeffcient in double format.
 */

double get_0th(PATTERN fromdct, int vector)
{
  return GetMatrixElement(fromdct->Data, 1, vector);
}


/* Calculates mean of all 0th coefficients in Cepstrum.
 * Takes as parameters Cepstrum in PATTERN data structure.
 * Returns mean in double format.
 */

double mean_0th(PATTERN fromdct)
{
  int i;
  double totalenergy;

  for (i = 1; i < PNumVectors(fromdct); i++) {
    totalenergy += get_0th(fromdct, i);
  }

  return totalenergy/PNumVectors(fromdct);
}

/* Calculates how many vectors are above threshold.
 * Takes as parameters Cepstrum in PATTERN data structure and threshold.
 * Returns number of vectors. 
 */

int above_threshold(PATTERN fromdct, double threshold)
{
  int i, nonsilent = 0;
  
  for (i = 1; i < PNumVectors(fromdct); i++) {
    if (get_0th(fromdct, i)> threshold)
      nonsilent++;
  }

  return nonsilent;
} 
