/* featextract.c by Ville Hautamki <villeh@cs.joensuu.fi>, 2002
   All sources are under GPL */

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h> // Needed for gettimeofday()
#include <time.h>
#include <unistd.h>   // Needed for getopt() 
#include <string.h>
#include <errno.h>

#include "util.h"
#include "speakerid.h"
#include "cmdline.h"

char *id = "adduser: Sat Jun  8 22:07:11 EET DST 2002";

long setwatch();
void prtime(long start);
Parameters *init_params(void);
void print_parameters(Parameters *parap);
void print_usage(); 
void free_parameters(Parameters *parap);
void cp_params(Parameters *parap, struct gengetopt_args_info *args_info);

int main(int argc, char **argv)
{
  long watch;
  Parameters *parap;
  struct gengetopt_args_info args_info;

  if (cmdline_parser(argc, argv, &args_info) != 0)
     exit(1) ;

  watch = setwatch();
  parap = init_params();
  cp_params(parap, &args_info);

  print_parameters(parap);

  if (!parap->featurefile) {
    printf("No outputfile!\n");
    exit(-1);
  }
    
  if (print_features(parap)) {
    printf("Failure in feature extract\n");
    exit(-1);
  }
    
  prtime(watch);

  free_parameters(parap);

  return 0;
}


/* Initializes time (for calculating running time).
 * Does not take any parameters.
 * Returns time (in seconds) See man 2 time
 */

long setwatch()
{
  time_t tmp;
  return (long) time(&tmp);
}

/* Prints running time. 
 * Takes as a parameter initialized time (from setwatch())
 * Does not return anything.
 */

void prtime(long start)
{
  struct timeval tv;
  double elapsed;

  gettimeofday(&tv, NULL);
  elapsed = (double)tv.tv_sec + ((double)tv.tv_usec /1e6) - start;

  printf("Time: %6.1f\n", elapsed);
}

/* Allocates memory and initializes Parameter struct.
 * Does not take any parameters.
 * Returns pointer to Parameters struct.
 */

Parameters *init_params(void)
{
  Parameters *parap;

  parap = (Parameters *) allocate(sizeof(Parameters), NO_INIT);

  parap->instr = NULL;
  parap->speaker_db = NULL;
  parap->path_models = NULL;
  parap->featptr = NULL;
  parap->modelptr = NULL;
  parap->matchptr = NULL;
  parap->deciptr = NULL;

  return parap;
}

/* Prints out parameters.
 * Takes a parameter pointer to Parameters struct.
 * Does not return anything.
 */

void print_parameters(Parameters *parap)
{
  printf("%s\n", id);
  printf("By Ville Hautamki, Teemu Kilpelinen, Tomi Kinnunen, Ismo Krkkinen\n");

  printf("Inputfile: %s\n", parap->instr);
  printf("Outputfile: %s\n", parap->featurefile);
  // Global parameters for whole signal
  printf("Window len: %d\n", parap->featptr->win_length);
  printf("Window shift: %d\n", parap->featptr->win_shift_ms);
  printf("Remove sil: %d\n", parap->featptr->remove_sil);
  printf("Silence Threshold Coeff: %f\n", parap->featptr->sil_thr);
  printf("Remove DC: %d\n", parap->featptr->remove_dc_offset);
  printf("Window function: %d\n", parap->featptr->win_function_type);
  printf("High emph coeff: %f\n", parap->featptr->high_emph_coeff);
  // Cepstral coeffient parameters
  printf("Feature type: %d\n", parap->featptr->feat_type);
  printf("Num. Mel filters: %d\n", parap->featptr->num_mel_filters);
  printf("Include delta: %d\n", parap->featptr->include_delta);
  printf("Include delta delta: %d\n", parap->featptr->include_delta_delta);
  printf("Include 0th coeff: %d\n", parap->featptr->include_0th_coeff);
  printf("MFCC ord: %d\n", parap->featptr->MFCC_ord);
  printf("Lifter type: %d\n", parap->featptr->lifter_type);
  printf("Absolute magnitude: %d\n", parap->featptr->absolute_magnitude);
  //Modeling parameters
  printf("Model size: %d\n", parap->modelptr->model_size);
  printf("Max iterations: %d\n", parap->modelptr->max_iterations);
  printf("Stop threshold: %f\n", parap->modelptr->stop_threshold);
  printf("Num. initial sol.: %d\n", parap->modelptr->num_initial_solutions);
  //Matching parameters
  printf("Matching func: %d\n", parap->matchptr->matching_function);
  printf("Using CB in matching: %d\n", parap->matchptr->using_cb);
  printf("Max iterations in CB matching: %d\n", parap->matchptr->maxiter);
  //Decision parameters
  printf("Decision func: %d\n", parap->deciptr->decision_function);
  printf("Number of best matches: %d\n", parap->deciptr->num_best);
  printf("Confidence measure: %d\n", parap->deciptr->confidence_measure);
}

/* Frees Parameters struct (and relating to it).
 * Takes as a parameters pointer to Parameters struct.
 * Does not return anything.
 */

void free_parameters(Parameters *parap)
{
  free(parap->deciptr);
  free(parap->matchptr);
  free(parap->modelptr);
  free(parap->featptr);
  free(parap->path_models);
  free(parap->speaker_db);
  free(parap->instr);
  free(parap);
}

/* Copies commandline parameters from struct gengetopt_args_info to 
 *  Parameters struct.
 * Takes as parameters pointer to Parameters struct and pointer to 
 *  gengetopt_args_info struct.
 * Does not return anything.
 */

void cp_params(Parameters *parap, struct gengetopt_args_info *args_info)
{

  parap->featptr = (Featextract *) allocate(sizeof(Featextract),NO_INIT);
  parap->modelptr = (Modeling *) allocate(sizeof(Modeling),NO_INIT);
  parap->matchptr = (Matching *) allocate(sizeof(Matching),NO_INIT);
  parap->deciptr = (Decisionlogic *) allocate(sizeof(Decisionlogic),NO_INIT);

  //Basic parameters
  parap->instr = args_info->input_arg;
  parap->speaker_db = args_info->speakerdb_arg;
  //  parap->path_models = args_info->modelpath_arg;
  parap->featurefile = args_info->featurefile_arg;
  // Global parameters for whole signal
  parap->featptr->win_length = args_info->win_length_arg;
  parap->featptr->win_shift_ms = args_info->win_shift_ms_arg;
  parap->featptr->remove_sil = args_info->remove_sil_flag;
  parap->featptr->sil_thr = args_info->sil_thr_arg;
  parap->featptr->remove_dc_offset = args_info->remove_dc_offset_flag;
  parap->featptr->win_function_type = args_info->win_function_type_arg;
  parap->featptr->high_emph_coeff = args_info->high_emph_coeff_arg;
  // Cepstral coeffient parameters
  parap->featptr->feat_type = args_info->feat_type_arg;
  parap->featptr->num_mel_filters = args_info->num_mel_filters_arg;
  parap->featptr->include_delta = args_info->include_delta_flag;
  parap->featptr->include_delta_delta = args_info->include_delta_delta_flag;
  parap->featptr->include_0th_coeff = args_info->include_0th_coeff_flag;
  parap->featptr->MFCC_ord = args_info->MFCC_ord_arg;
  parap->featptr->lifter_type = args_info->lifter_type_arg;
  parap->featptr->absolute_magnitude = args_info->absolute_magnitude_flag;
}
