/* ---------------------------------------------------------------------*/
/*  feature_extraction.c   Ville Hautamki <villeh@cs.joensuu.fi>, 2002 */ 
/*                         Teemu Kilpelinen 17.05.2001		        */
/*   									*/
/*  This module is responsible of feature extraction                    */
/* ---------------------------------------------------------------------*/

#include <float.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fftw.h>

#include "pattern.h"
#include "matrix.h"
#include "signal.h"
#include "speakerid.h"  //for parameter struct
#include "util.h"
#include "frame.h"
#include "dct.h"
#include "mel-functions.h"
#include "feature_extraction.h"

/* Does the feature extraction. 
 * Takes as parameteres pointer to Signal struct and pointer to Parameters
 *  struct. 
 * Returns Cepstrum feature table in PATTERN data structure.
 */

PATTERN FeatureExtraction(Signal *sigptr, Parameters *parap) 
{ 
  int count_of_vectors; //i; For bedugging
  Frame *frptr;
  double *mel_filter_bank = NULL, *log_filtered_spectrum, *lifter_function;
  Filterbank *filter_bank;
  fftw_complex *out, *in;
  fftw_plan fftplan;
  PATTERN dct, cepstrum;
  MATRIX vector;


  // Initialize and construct mel filter bank.
  frptr = init_frame(sigptr, parap);
  fftplan = fftw_create_plan(frptr->length, FFTW_FORWARD, FFTW_ESTIMATE);
  if (parap->featptr->num_mel_filters == 0)
    parap->featptr->num_mel_filters = NumberOfMelFilters(sigptr->srate);
  filter_bank = CreateFilterBank(parap->featptr->num_mel_filters, 
				 frptr->length, sigptr->srate);
  lifter_function = LifteringCoefficients(parap->featptr->lifter_type, 
					  parap->featptr->MFCC_ord);
  count_of_vectors = (int)(sigptr->num_samples/
			   (double)frptr->win_shift_samples);
  out = (fftw_complex *) allocate(sizeof(fftw_complex)*frptr->length, INIT); 
  in = (fftw_complex *) allocate(sizeof(fftw_complex)*frptr->length, INIT); 
  log_filtered_spectrum = (double *) allocate(sizeof(double)*
					     parap->featptr->num_mel_filters, 
					     NO_INIT);
  dct = NewPattern(parap->featptr->MFCC_ord + 1, count_of_vectors);
  count_of_vectors = 0;

  // Loop where is implemented the handling of frames and FFT itself
  while (frptr->begin + frptr->samples_per_frame < sigptr->num_samples) {
    MakeFrame(sigptr, frptr);

    copy_frame_complex(in, frptr);
    zero_arr(frptr->length, out);
    fftw_one(fftplan, in, out);
    copy_complex_frame(out, frptr, parap->featptr->absolute_magnitude);

    ComputeLogFilteredSpectrum(frptr->frame, log_filtered_spectrum, 
			       filter_bank, parap);    

    vector = DCT(log_filtered_spectrum, parap); 

    // Not liftering if lifter_type == 1 (truncated)
    if (parap->featptr->lifter_type != 1)
      CepstralLiftering(vector, lifter_function, parap->featptr->MFCC_ord);

    count_of_vectors++;	
    SetVector(dct, vector, count_of_vectors);
    frptr->begin += frptr->win_shift_samples;
  }

  cepstrum = remove_silent_frames(dct, parap);	
  if ((parap->featptr->remove_sil == 0) && 
      (parap->featptr->include_0th_coeff == 1))
    ;
  else // Otherwise you have to destroy old cepstrum
    DeletePattern(dct);

  /* memory deallocations */
  fftw_destroy_plan(fftplan);
  free_frame(frptr);
  free_filterbank(filter_bank);
  free(out);
  free(in);
  free(log_filtered_spectrum);
  free(mel_filter_bank);
  free(lifter_function);

  return cepstrum;
}

/* Implements the Cepstral Liftering.
 * Takes as parameters one vector in MATRIX data structure, lifter function in 
 *  double array and lifter length (integer).
 * Does not return anything.
 */

void CepstralLiftering(MATRIX vector, double *lifter_function, 
		       int lifter_length)
{
 int k;
 double tmp; 
        
 for (k = 1; k <= lifter_length; k++)
   tmp = GetMatrixElement(vector, k, 1);
   SetMatrixElement(vector, k, 1, tmp * lifter_function[k - 1]);
}

/* Calculates Liftering coeffs
 * Takes as parameters lifter type (integer) and lifter length (integer).
 * Returns pointer to lifter function, which is double array.
 */

double *LifteringCoefficients(int lifter_type, int lifter_length)
{
  int k;
  double h, *lifter_function;	
	
  lifter_function = (double *) allocate(sizeof(double)*lifter_length, NO_INIT);
	
  /* raised sine */	
  if (lifter_type == 2){	
    h = lifter_length / 2.0f;
    for (k = 1; k <= lifter_length; k++)
      lifter_function[k - 1] = (double)(1 + (h * 
					    sin((k * PI)/lifter_length))); 
  } else { /* half raised sine */
    for (k = 1; k <= lifter_length; k++)
      lifter_function[k - 1] = 
	(double)(1 + (lifter_length/2 * sin(1/2.0 * ((k * PI)/lifter_length))));
  }

  return lifter_function;
}

/* Copies Real frame to FFTW Complex number data structure.
 * Takes as parameters pointer to FFTW Complex data structure, Pointer to Frame
 *  struct. 
 * Does not return anything.
 */

void copy_frame_complex(fftw_complex *in, Frame *frptr)
{
  int i;

  for (i = 0; i < frptr->length; i++) {
    c_re(in[i]) = frptr->frame[i];
    c_im(in[i]) = 0.0;
  }
}

/* Zeroes FFTW Complex data structure. 
 * Takes as parameters frame length and pointer to FFTW Complex data structure
 * Does not return anything.
 */

void zero_arr(int n, fftw_complex *a)
{
     int i;

     for (i = 0; i < n; ++i)
          c_re(a[i]) = c_im(a[i]) = 0.0;
}

/* This function calculates complex abs
 * in two possible ways: 
 * Power spectrum (POWERSPECTR) or 
 * Magnitude spectrum (MAGNITUDESPECTR)
 * 
 *Parameters: pointer to fftw_complex frame, pointer to output frame and
 *            flag to choose method.
 *
 *Returns:    as a parameter frame.
 */

void copy_complex_frame(fftw_complex *in, Frame *frptr, int absolute_magnitude)
{
  int i;

  if (absolute_magnitude == MAGNITUDESPECTR) {
    for (i = 0; i < frptr->length; i++) 
      frptr->frame[i] = sqrt(c_re(in[i]) * c_re(in[i]) + c_im(in[i]) * c_im(in[i]) ) + DBL_EPSILON;
  } else {
    for (i = 0; i < frptr->length; i++) 
      frptr->frame[i] = c_re(in[i]) * c_re(in[i]) + c_im(in[i]) * c_im(in[i]) 
	+ DBL_EPSILON;
  }
}
