/* ----------------------------------------------------------------------*/
/*  frame.c Ville Hautamki, <villeh@cs.joensuu.fi>, 2002                */
/*          Teemu Kilpelinen 20.10.2000		                 */
/*	This module includes MakeFrame function for frame handling.	 */  
/*	Version Number: 0.01						 */
/* ----------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "signal.h"
#include "speakerid.h" 
#include "util.h"
#include "frame.h"

/* Inits frame data structure.
 * Takes as parameters pointer to signal struct and pointer to Paramters struct
 * Returns pointer to Frame struct.
 */

Frame *init_frame(Signal *sigptr, Parameters *parap) 
{
  Frame *frptr;
  
  frptr = (Frame *) allocate(sizeof(Frame), NO_INIT);

  frptr->samples_per_frame = Ms2Samples(parap->featptr->win_length, 
					get_srate(sigptr));
  frptr->length = NearestPowerOfTwo(frptr->samples_per_frame);
  frptr->begin = 0;
  frptr->win_shift_samples = Ms2Samples(parap->featptr->win_shift_ms, 
					get_srate(sigptr));
  frptr->frame = (double *) allocate(sizeof(double)*frptr->length, INIT); 
  frptr->win_coeff = (double *) allocate(sizeof(double)*
					 frptr->samples_per_frame, NO_INIT); 

  WindowCoefficients(frptr, parap->featptr->win_function_type);
  
  return frptr;
}

/* Frees frames.
 * Takes as a parameters pointer to Frame struct.
 * Does not return anything.
 */

void free_frame(Frame *frptr)
{
  free(frptr->frame);
  free(frptr->win_coeff);
  free(frptr);
}

/*
	Functioning:	Compute coefficients for time domain window function.  
	
	Inputs:	      Pointer to Frame struct.
		      int window_type		  1 = Hamming, 2 = Barlett, 
					          3 = Hanning, 4 = Blackman
		      						
	Outputs:		Time domain window function.
	
	Return value:	-
                                Ville Hautamki, 2002 
	History:		Teemu Kilpelinen 15.04.2001.
	Version:		1.1 

	Reference:		
*/

void WindowCoefficients(Frame *frptr, int window_type)
{
  int k, window_length;

  window_length = frptr->samples_per_frame;

  for(k = 0; k < window_length; k++) { 
    if (window_type == HAMMING)   
      frptr->win_coeff[k] = (double)(0.54 - 0.46 * 
				   cos(2 * PI * k /window_length));	

    else if (window_type == BARLETT) {
      if (k <= (window_length / 2.0))
	frptr->win_coeff[k] = (2.0f * k) / window_length;
      else frptr->win_coeff[k] = 2.0f - (2.0f * k)/window_length;

    } else if (window_type == HANNING) 
      frptr->win_coeff[k] = (double)(0.5 - 0.5 * 
				    cos(2 * PI * k / window_length));

    else /* BLACKMAN */
      frptr->win_coeff[k] = (double)(0.42 - 0.5 * 
				   cos(2 * PI * k / window_length) + 0.08 * 
				   cos(4 * PI * k / window_length));
  }
} 



/* --------------------------------------------------------------- */
/*
	Functioning:	Make frame     Copies samples and multiply them with 
				       window coefficients (Hamming, etc.)
				       from input signal to frame which length 
				       is power of two.
	Inputs:	   Pointer to Signal struct and pointer to Frame struct.

        Outputs:    A frame which length is power of two (it can have nulls at 
					the begin and the end).    
	Return value:	- 
                          Ville Hautamki, 2002 
	History:	  Teemu Kilpelinen and Tomi Kinnunen	 25.1.2001
	Version:		1.1

	Reference:		
*/

void MakeFrame(Signal *sigptr, Frame *frptr)
{
  int k;
	
  /*j is variable, which in is the index where from the filling begins 
    (if used not use round) downwards. */
	
  int j = (int)((frptr->length - frptr->samples_per_frame) / 2.0); 

  memset(frptr->frame, 0, sizeof(double) * frptr->length);	

  for(k = 0; k < frptr->samples_per_frame; k++)
    frptr->frame[k+j] = frptr->win_coeff[k] * 
      sigptr->signal_buffer[k + frptr->begin];
} 

/* Prints out a frame. (used for debugging)
 * Takes as parameters pointer to Frames struct and Pointer to opened file.
 * Does not return anything.
 */

void print_frame(Frame *frptr, FILE *output)
{
  int i;

  for (i = 0; i < frptr->length; i++)
    fprintf(output, "%f ", frptr->frame[i]);

  fprintf(output,"\n\n");
}






