/*******************************************************************

    FVEC.C

    22.5.2001  iak

    Version 0.03

    Operations for double vectors and sets of double vectors.

*******************************************************************/

#include <stdlib.h>

#include "fvec.h"
#include "random.h"
#include "textfile.h"
#include "util.h"

void fvCopy(double* From, double* To, int Dimensionality) {
    while (Dimensionality--) *To++ = *From++;
}

void fvFill(double* V, double Value, int Dimensionality) {
    while (Dimensionality--) *V++ = Value;
}

void fvRandom(double* V, int Dimensionality) {
    while (Dimensionality--) *V++ = frand();
}

double fvManhattanDistance(double* V1, double* V2, int Dimensionality) {
    double Dist = 0;
    while (Dimensionality--) Dist += fabs(*V1++ - *V2++);
    return Dist;
}

double fvManhattanLength(double* V, int Dimensionality) {
    double Len = 0;
    while (Dimensionality--) Len += fabs(*V++);
    return Len;
}

double fvDistance(double* V1, double* V2, int Dimensionality) {
    double Dist;
    double* tmp = fvNew(Dimensionality);
    fvSubtract(V1, V2, tmp, Dimensionality);
    Dist = fvLength(tmp, Dimensionality);
    fvDelete(tmp);
    return Dist;
}

double fvScalarProduct(double* V1, double* V2, int Dimensionality) {
    double Prod = 0;
    while (Dimensionality--) Prod += *V1++ * *V2++;
    return Prod;
}

void fvAdd(double* V1, double* V2, double* Result, int Dimensionality) {
    while (Dimensionality--) *Result++ = *V1++ + *V2++;
}

void fvSubtract(double* V1, double* V2, double* Result, int Dimensionality) {
    while (Dimensionality--) *Result++ = *V1++ - *V2++;
}

void fvScale(double* V, double Scale, int Dimensionality) {
    while (Dimensionality--) {
	*V = *V * Scale;
	V++;
    }
}

static int fv_asc = 0;

static int double_compare(const void* a, const void* b) {
    double aa, bb;
    aa = *(double*)a;
    bb = *(double*)b;
    if (fv_asc) {
        if (aa < bb) return -1;
        if (aa > bb) return 1;
    } else {
        if (aa < bb) return 1;
        if (aa > bb) return -1;
    }
    return 0;
}

void fvSort(double* V, int Dimensionality, int Ascending) {
    fv_asc = Ascending;
    qsort(V, Dimensionality, sizeof(double), double_compare);
}

double** fvNewSet(int Count, int Dimensionality) {
    int k;
    double** Set = allocate(sizeof(double*) * Count, NO_INIT);
    for (k = 0; k < Count; k++) Set[k] = fvNew(Dimensionality);
    return Set;
}

void fvDeleteSet(double** Set, int Count) {
    while (Count--) fvDelete(Set[Count]);
    free(Set);
}

double** fvCopySet(double** Dest, double** Src, int Count, int Dimensionality) {
    int k;
    for (k = 0; k < Count; k++) fvCopy(Src[k], Dest[k], Dimensionality);
    return Dest;
}

int fvWriteSet(double** Set, int Count, int Dimensionality, FILE* f) {
    int k, n, e;
    for (n = 0; n < Count; n++) {
	for (k = 0; k < Dimensionality - 1; k++)
	    fprintf(f, "%e ", Set[n][k]);
	fprintf(f, "%e\n", Set[n][Dimensionality - 1]);
	e = ferror(f);
	if (e) return e;
    }
    return 0;
}

static void Convert(double* DV, double* FV, int Dim) {
    while (Dim--) *FV++ = (double) *DV++;
}

int fvReadSet(double*** Set, int* Count, int* Dimensionality, FILE* f) {
    double* Vec;
    int Incomplete, Errors, D, Incomp, Err, Stage, RetVal = 1;
    fpos_t Beginning;
    fgetpos(f, &Beginning);
    *Set = 0;
    *Dimensionality = 0;
    LineNumber = 0;
    for (Stage = 0; Stage < 2; Stage++) {
	*Count = 0;
        if (!feof(f)) Vec =
	    ReadRawVector(f, Dimensionality, &Incomplete, &Errors);
	else return 0;
	if (Stage && Vec) Convert(Vec, (*Set)[0], *Dimensionality);
        if (Errors) {
            fprintf(stderr, "Error in line: %i\n", LineNumber);
            return 0;
        }
	*Count = 1;
	D = 0;
        while (!feof(f)) {
    	    Vec = ReadRawVector(f, &D, &Incomp, &Err);
	    if (Err) {
                fprintf(stderr, "Error in line: %i\n", LineNumber);
                return 0;
            }
	    if (Vec) {
	        if (D != *Dimensionality) {
		    RetVal = -1;
                    if (!Stage) 
		       fprintf(stderr, "WARNING: Dimensionality mismatch in line %i, line ignored\n",
                               LineNumber);
                } 
		else {
	           if (Stage) Convert(Vec, (*Set)[*Count], *Dimensionality);
	           (*Count)++;
		}
	    }
        }
	if (!Stage) {
	    fsetpos(f, &Beginning);
	    *Set = fvNewSet(*Count, *Dimensionality);
	}
    }
    return RetVal;
}

void fvFillSet(double** Set, int Count, int Dimensionality, double Value) {
    int k;
    for (k = 0; k < Count; k++) fvFill(Set[k], Value, Dimensionality);
}

double** fvSubSetDimensions(double** Set, int Count, int Dimensionality,
    int* SelectedDimensions, double** SubSet, int* SubsetDimensionality)
{
    int k, n, Dims;
    Dims = 0;
    for (k = 0; k < Dimensionality; k++) if (SelectedDimensions[k]) Dims++;
    if (SubsetDimensionality) *SubsetDimensionality = Dims;
    if (!SubSet) SubSet = fvNewSet(Count, Dims);
    for (n = 0; n < Count; n++) {
	for (k = 0, Dims = 0; k < Dimensionality; ++k, ++Dims)
	    if (SelectedDimensions[k]) SubSet[n][Dims] = Set[n][k];
    }
    return SubSet;
}

double fvSetAverage(double** Set, int Count, int Dimension) {
    int k;
    double Sum = 0;
    for (k = 0; k < Count; k++) Sum += Set[k][Dimension];
    return Sum / Count;
}

double fvSetNth(double** Set, int Count, int Dimension, int N) {
    int k;
    double Result;
    double* Values = allocate(sizeof(double) * Count, NO_INIT);
    for (k = 0; k < Count; k++) Values[k] = Set[k][Dimension];
    fv_asc = 1;
    qsort(Values, Count, sizeof(double), &double_compare);
    Result = Values[N];
    free(Values);
    return Result;
}

double fvSetMinimum(double** Set, int Count, int Dimension) {
    int k;
    double Result = Set[0][Dimension];
    for (k = 1; k < Count; k++)
	if (Result > Set[k][Dimension]) Result = Set[k][Dimension];
    return Result;
}

double fvSetMaximum(double** Set, int Count, int Dimension) {
    int k;
    double Result = Set[0][Dimension];
    for (k = 1; k < Count; k++)
	if (Result < Set[k][Dimension]) Result = Set[k][Dimension];
    return Result;
}

double fvSetTotalMinimum(double** Set, int Count, int Dimensionality) {
    int k;
    double Minimum, tmp;
    Minimum = fvSetMinimum(Set, Count, 0);
    for (k = 1; k < Dimensionality; k++) {
	tmp = fvSetMinimum(Set, Count, k);
	if (tmp < Minimum) Minimum = tmp;
    }
    return Minimum;
}

double fvSetTotalMaximum(double** Set, int Count, int Dimensionality) {
    int k;
    double Maximum, tmp;
    Maximum = fvSetMaximum(Set, Count, 0);
    for (k = 1; k < Dimensionality; k++) {
	tmp = fvSetMaximum(Set, Count, k);
	if (tmp > Maximum) Maximum = tmp;
    }
    return Maximum;
}

int fvSetIsBinary(double** Set, int Count, int Dimensionality, 
    double V1, double V2)
{
    int k, n;
    for (n = 0; n < Count; n++) for (k = 0; k < Dimensionality; k++)
	if (Set[n][k] != V1 || Set[n][k] != V2) return 0;
    return 1;
}

static int fv_dim = 0;
static int* fv_dord = 0;
static int* fv_vord = 0;
static int fv_ord = 0;

static int fv_gen_compare(const void* aa, const void* bb) {
    int k;
    double* a = *(double**)aa;
    double* b = *(double**)bb;
    if (!fv_dord || !fv_dord) for (k = 0; k < fv_dim; k++) {
	if (fv_ord) {
	    if (a[k] < b[k]) return 1;
	    if (a[k] > b[k]) return -1;
	} else {
	    if (a[k] < b[k]) return -1;
	    if (a[k] > b[k]) return 1;
	}
    } else for (k = 0; k < fv_dim; k++) {
	if (fv_vord[k]) {
	    if (a[fv_dord[k]] < b[fv_dord[k]]) return 1;
	    if (a[fv_dord[k]] > b[fv_dord[k]]) return -1;
	} else {
	    if (a[fv_dord[k]] < b[fv_dord[k]]) return -1;
	    if (a[fv_dord[k]] > b[fv_dord[k]]) return 1;
	}
    }
    return 0;
}

void fvSortSet(double** Set, int Count, int Dimensionality,
    int* DimensionOrder, int* ValueOrder, int Order)
{
    fv_dord = DimensionOrder;
    fv_vord = ValueOrder;
    fv_dim = Dimensionality;
    fv_ord = Order;
    qsort(Set, Count, sizeof(double*), &fv_gen_compare);
}

double* fvSetUnique(double** Set, int Count, int Dimension, int* FoundCount) {
    int k, n, Sorted;
    double* List, *Result;
    if (!Count) {
	*FoundCount = 0;
	return 0;
    }
    Sorted = 1;
    List = allocate(sizeof(double) * Count, NO_INIT);
    List[0] = Set[0][Dimension];
    for (k = 1; k < Count; k++) {
	List[k] = Set[k][Dimension];
	if (List[k - 1] > List[k]) Sorted = 0;
    }
    if (!Sorted) fvSort(List, Count, 1);
    *FoundCount = 1;
    for (k = 1; k < Count; k++) if (List[k - 1] != List[k]) (*FoundCount)++;
    Result = allocate(sizeof(double) * *FoundCount, NO_INIT);
    k = 0;
    for (n = 0; n < *FoundCount; n++) {
	Result[n] = List[k++];
	while (k < Count && List[k] == List[k - 1]) k++;
    }
    free(List);
    return Result;
}
