/* ---------------------------------------------------------------------------------------------*/
/*                                                                                              */
/*                              matrix.c                                                        */
/*												*/
/*		Basic matrix manipulation operations.						*/  		
/*												*/
/*		See any numerical linear algebra book for more information			*/
/*		about the definitions etc. for these operations.				*/
/*                                                                                              */
/*              Copyright (C):                                                                  */
/*                      University of Joensuu, Department of Computer Science                   */
/*                                                                                              */
/*              Author(s) and current version:                                                  */
/*                                                                                              */
/*                      Tomi Kinnunen & Ismo Krkkinen                                         */
/*                      August 2001 - version 0.1  	                                        */
/*                      tkinnu@cs.joensuu.fi, iak@cs.joensuu.fi                                 */
/*                                                                                              */
/* -------------------------------------------------------------------------------------------- */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "matrix.h"
#include "misc.h"	

MATRIX NewMatrix(int Rows, int Cols) {
/* ---------------------------------------------------------------------------------------------*/
/*                                                                                              */
/*      Creates a new matrix of size Rows x Cols.                                               */
/*                                                                                              */
/* -------------------------------------------------------------------------------------------- */

	MATRIX M = malloc(sizeof(MATRIX_TYPE));	

	if ( (Rows < 0) || (Cols < 0) )
		SegFault("ERROR [NewMatrix]: Matrix size must be at least 0 x 0 !");
	else {
		M->Rows = Rows; 
		M->Cols = Cols;
		M->Data = malloc(Rows * Cols * sizeof(double));
	}
	
	return M;
} /* NewMatrix */

void DeleteMatrix(MATRIX M) {
/* ---------------------------------------------------------------------------------------------*/
/*                                                                                              */
/*      Deletes a matrix.			                                                */     
/*                                                                                              */
/* -------------------------------------------------------------------------------------------- */

	free(M->Data);
	free(M);

} /* DeleteMatrix */

void SetMatrixElement(MATRIX M, int Row, int Col, double Value) {
/* ---------------------------------------------------------------------------------------------*/
/*                                                                                              */
/*      Sets a value into matrix: M[Row,Col] = Value.                                           */     
/*                                                                                              */
/* -------------------------------------------------------------------------------------------- */

	int Pos;

	if ( (Row < 1) || (Col < 1) || (Row > MRows(M)) || (Col > MCols(M)) )
                SegFault("ERROR [SetMatrixElement]: Matrix subscript out of range!");
	
	Pos = MCols(M)*(Row - 1) + Col-1;  
	M->Data[Pos] = Value;	

} /* SetMatrixElement */

double GetMatrixElement(MATRIX M, int Row, int Col) {
/* ---------------------------------------------------------------------------------------------*/
/*                                                                                              */
/*      Retrieves a matrix element from location M[Row,Col].                                    */
/*                                                                                              */
/* -------------------------------------------------------------------------------------------- */

	int Pos;

	if ( (Row < 1) || (Col < 1) || (Row > MRows(M)) || (Col > MCols(M)) )
                SegFault("ERROR [GetMatrixElement]: Matrix subscript out of range!");
	
	Pos = MCols(M)*(Row - 1) + Col-1;

	return M->Data[Pos];
} /* GetMatrixElement */

MATRIX GetRow(MATRIX TargetVector, MATRIX M, int Row) {

	int j;
	
	if ( (Row < 1) || (Row > MRows(M)) )
                SegFault("ERROR [GetRow]: Matrix subscript out of range!");
	else if (!(IsRowVector(TargetVector)))
		SegFault("ERROR [GetRow]: Target vector is not row vector!");

	for(j=1; j <= MCols(M); j++)
		SetMatrixElement(TargetVector,1,j,GetMatrixElement(M,Row,j));

	return TargetVector;

}

MATRIX GetColumn(MATRIX TargetVector, MATRIX M, int Col) {
 
	int i;
        
        if ( (Col < 1) || (Col > MCols(M)) )                                  
                SegFault("ERROR [GetColumn]: Matrix subscript out of range!");
	else if (!(IsColVector(TargetVector)))
		SegFault("ERROR [GetColumn]: Target vector is not column vector!");
        
        for(i=1; i <= MRows(M); i++)  
                SetMatrixElement(TargetVector,i,1,GetMatrixElement(M,i,Col));
                
	return TargetVector;
	
}

MATRIX SetRow(MATRIX M, MATRIX Data, int Row) {

	int i;

	if ( (Row < 1) || (Row > MRows(M)) )
                SegFault("ERROR [SetRow]: Matrix subscript out of range!");
	else if ( MCols(Data) != MCols(M) )
		SegFault("ERROR [SetRow]: Mismatched number of columns in matrices!");
	else if ( MRows(Data) > 1 )
		SegFault("ERROR [SetRow]: Matrix must have only one row!");

	for(i=1; i <= MCols(M); i++)
                SetMatrixElement(M,Row,i,GetMatrixElement(Data,1,i));

	return M;

}

MATRIX Diag(MATRIX TargetVector, MATRIX M) {

	int i;

	if ( MRows(TargetVector) != MRows(M) )
		SegFault("ERROR [DiagonalElements] : Dimension mismatch.");
	
	for(i=1; i <= MRows(M); i++)
		SetMatrixElement(TargetVector, i, 1, 
			GetMatrixElement(M, i, i));

	return TargetVector;
}

double MaxElem( MATRIX M ) {
   int   i, j;
   double Max, Val;

   if (MRows(M) == 0 && MCols(M) == 0)
      SegFault("ERROR [MaxElem]: non-empty matrix expected.");

   Max = GetMatrixElement(M, 1, 1);
   for(i=1; i <= MRows(M); i++)
      for (j=1; j <= MCols(M); j++) {
         Val = GetMatrixElement(M, i, j);
	 if ( Val > Max ) 
	    Max = Val;
      }
	
   return Max;

}

double MinElem( MATRIX M ) {
   int   i, j;
   double Min, Val;
        
   if (MRows(M) == 0 && MCols(M) == 0)
      SegFault("ERROR [MaxElem]: non-empty matrix expected.");
                
   Min = GetMatrixElement(M, 1, 1);
   for(i=1; i <= MRows(M); i++)     
      for (j=1; j <= MCols(M); j++) {
         Val = GetMatrixElement(M, i, j);
         if ( Val < Min ) 
            Min = Val;
      }
        
   return Min;   

}


MATRIX SetColumn(MATRIX M, MATRIX Data, int Col) {

	int i;
 
        if ( (Col < 1) || (Col > MCols(M)) )
                SegFault("ERROR [SetColumn]: Matrix subscript out of range!");
        else if ( MRows(Data) != MRows(M) )
                SegFault("ERROR [SetColumn]: Mismatched number of rows in matrices!");
        else if ( MCols(Data) > 1 )
                SegFault("ERROR [SetColumn]: Matrix must have only one column!");
        
        for(i=1; i <= MRows(M); i++)
                SetMatrixElement(M,i,Col,GetMatrixElement(Data,i,1));  
 
        return M;

}

MATRIX InsertNewVectorToMatrix(MATRIX M, MATRIX Vector, int ColVec) {
/* ColVec == 1 --> insert a column vector. Otherwise row vector */

	double*	Src;
	double*	Dst;
	int 	i,j;

	if ( (ColVec) && (MRows(Vector) != MRows(M)))
		SegFault("ERROR [InsertNewVector]: Mismatched number of rows!");
	else if ( !(ColVec) && (MCols(Vector) != MCols(M)))
		SegFault("ERROR [InsertNewVector]: Mismatched number of columns!");
	else if ( !(ColVec) ) {
		M->Data = realloc(M->Data, sizeof(double)*(MRows(M)+1)*(MCols(M)) );
		M->Rows += 1;
		SetRow(M, Vector, MRows(M));
	}
	else if (ColVec) {
		M->Data = realloc(M->Data, sizeof(double)*(MCols(M)+1)*(MRows(M)) );
		Src = M->Data + MCols(M) * MRows(M) - 1;
		Dst = M->Data + (MCols(M)+1) * MRows(M) - 1;
		for(i=1; i <= MRows(M); i++) {
			Dst--;
			for(j=1; j <= MCols(M); j++)
				*Dst-- = *Src--;
		}
		M->Cols += 1;		  
                SetColumn(M, Vector, MCols(M));
	}
	else
		SegFault("ERROR [InsertNewVector]: Don't know how to insert this vector!");

	return M;
}

MATRIX Transpose(MATRIX M) {
/* ---------------------------------------------------------------------------------------------*/
/*                                                                                              */
/*	Matrix transposition - i.e. changer rows to columns and vice versa.			*/
/*                                                                                              */
/* -------------------------------------------------------------------------------------------- */

	int	i, j, Pos1, Pos2;

	double* 	fooMatrix = malloc(sizeof(double) * MRows(M) * MCols(M));
	
	for(i=1; i <= MRows(M); i++)
		for(j=1; j <= MCols(M); j++) {
			Pos1 = MCols(M)*(i-1) + j-1;
			Pos2 = MRows(M)*(j-1) + i-1;
			fooMatrix[Pos2] = M->Data[Pos1];
		}
	
	free(M->Data);
	M->Data = fooMatrix;

	i = M->Rows;
	M->Rows = M->Cols;
	M->Cols = i;

	return M;

} /* Transpose */

void PrintMatrix(MATRIX M, FILE* f) {
/* ---------------------------------------------------------------------------------------------*/
/*                                                                                              */
/*	Prints a matrix.									*/
/*	For example, you might want to write PrintMatrix(A, stderr);				*/
/*                                                                                              */
/* -------------------------------------------------------------------------------------------- */

	int i, j;

	for(i=1; i <= MRows(M); i++) {
                for(j=1; j <= MCols(M); j++)
			fprintf(f, "%5.5f ", GetMatrixElement(M,i,j));
		fprintf(f, "\n");  
	}
} /* PrintMatrix */

MATRIX IdentityMatrix(MATRIX M) {
/* ---------------------------------------------------------------------------------------------*/
/*                                                                                              */
/*	Creates a new matrix with ones in the diagonal and zeros elsewhere.			*/
/*                                                                                              */
/* -------------------------------------------------------------------------------------------- */

	int 	i, j;

	for(i=1; i <= MRows(M); i++)
		for(j=1; j <= MCols(M); j++)
			if (i == j)
				SetMatrixElement(M, i, j, 1);
			else
				SetMatrixElement(M, i, j, 0);

	return M;
	
} /* IdentityMatrix */

MATRIX ZeroMatrix(MATRIX M) {
/* ---------------------------------------------------------------------------------------------*/
/*                                                                                              */
/*	Creates a matrix with zero elements.							*/
/*                                                                                              */
/* -------------------------------------------------------------------------------------------- */

	int     i, j;
                
        for(i=1; i <= MRows(M); i++) 
                for(j=1; j <= MCols(M); j++)
                	SetMatrixElement(M, i, j, 0);
        
	return M;
                
} /* ZeroMatrix */

MATRIX RandomMatrix(MATRIX M) {
/* ---------------------------------------------------------------------------------------------*/
/*                                                                                              */
/*	Creates a random matrix with values in the range [0, 1].				*/
/*	Remember to initialize random number generator with srand() !				*/
/*	An example:										*/
/*		#include <unistd.h>		// getpid() here				*/
/*		// ...										*/
/*		srand(getpid());								*/
/*		MATRIX A = RandomMatrix(3,3);							*/
/*		// ...										*/
/*                                                                                              */
/* -------------------------------------------------------------------------------------------- */

	int     i, j;
        
        for(i=1; i <= MRows(M); i++)
                for(j=1; j <= MCols(M); j++)
                        SetMatrixElement(M, i, j, rand()/(double)RAND_MAX);

	return M;
                        
} /* RandomMatrix */

MATRIX MultiplyByScalar(MATRIX Result, MATRIX M, double Scalar) {
/* ---------------------------------------------------------------------------------------------*/
/*                                                                                              */
/*	Multiplies each matrix M element by Scalar.						*/
/*                                                                                              */
/* -------------------------------------------------------------------------------------------- */

	int     i, j;

	if ((MRows(Result) != MRows(M)) || (MCols(Result) != MCols(M)))
		SegFault("ERROR [MultiplyByScalar]: Result and target matrices don't agree \
			with dimensions!");

	for(i=1; i <= MRows(M); i++)
                for(j=1; j <= MCols(M); j++)
                	SetMatrixElement(Result, i, j, Scalar * GetMatrixElement(M, i, j) );                

	return Result;

} /* MultiplyByScalar */

MATRIX MatrixSum(MATRIX Result, MATRIX A, MATRIX B) {
/* ---------------------------------------------------------------------------------------------*/
/*                                                                                              */
/*	Sum of two matrices.									*/
/*                                                                                              */
/* -------------------------------------------------------------------------------------------- */

	int	i, j;
	double 	foo;

	if ( (MRows(A) != MRows(B)) || (MCols(A) != MCols(B)) )
		SegFault("ERROR [MatrixSum]: matrix sizes don't agree.");

	for(i=1; i <= MRows(A); i++)
		for(j=1; j <= MCols(A); j++) {
			foo = GetMatrixElement(A, i, j) 
				+ GetMatrixElement(B, i, j);
			SetMatrixElement(Result, i, j, foo); 
		}

	return Result;

} /* MatrixSum */

MATRIX MatrixSubstract(MATRIX Result, MATRIX A, MATRIX B) {
/* ---------------------------------------------------------------------------------------------*/
/*                                                                                              */
/*	Substracts each element of matrix B from matrix A, i.e. A - B				*/
/*                                                                                              */
/* -------------------------------------------------------------------------------------------- */

	int     i, j;
        double   foo;    
        
        if ( (MRows(A) != MRows(B)) || (MCols(A) != MCols(B)) )
                SegFault("ERROR [MatrixSubstract]: matrix sizes don't agree.");

        for(i=1; i <= MRows(A); i++)
                for(j=1; j <= MCols(A); j++) {
                        foo = GetMatrixElement(A, i, j)
                                - GetMatrixElement(B, i, j);
                        SetMatrixElement(Result, i, j, foo);
                }
	
	return Result;      
  
} /* MatrixSubstract */

MATRIX MatrixProduct(MATRIX Result, MATRIX A, MATRIX B) {
/* ---------------------------------------------------------------------------------------------*/
/*                                                                                              */
/*	Matrix product.										*/
/*												*/
/*	Remember: inner dimensions must agree, i.e. if A = [a(ij)] M x N			*/
/*		  then matrix B must be B = [b(ij)] N x P					*/	      	
/*                                                                                              */
/* -------------------------------------------------------------------------------------------- */

	int 	i, j, k;
	double 	sum;	

	if ( MCols(A) != MRows(B) )
		SegFault("ERROR [MatrixProduct]: Failed to compute matrix product; inner \
			dimensions don't agree!");
	if ( MRows(Result) != MRows(A) )
		SegFault("ERROR [MatrixProduct]: Dimension mismatch in result and target matrices");
	
	for(i=1; i <= MRows(A); i++)
		for(j=1; j <= MCols(B); j++) {
			sum = 0;
			for(k=1; k <= MCols(A); k++)
				sum += GetMatrixElement(A, i, k)	
					* GetMatrixElement(B, k, j);
			SetMatrixElement(Result, i, j, sum);	
		}

	return Result;

} /* MatrixProduct */

static MATRIX MatrixExcept(MATRIX M, int Row, int Col) {
/* ---------------------------------------------------------------------------------------------*/
/*                                                                                              */
/*	Returns a sub-matrix of M where row Row and column Col are removed.			*/ 
/*	(A static help function needed in the calculation of determinant).			*/
/*                                                                                              */
/* -------------------------------------------------------------------------------------------- */

	int i, j;

	MATRIX C = NewMatrix(MRows(M)-1, MCols(M)-1);

	/* Left upper corner */
	for(i=1; i <= Row-1; i++)
		for(j=1; j <= Col-1; j++)
			SetMatrixElement(C, i, j, GetMatrixElement(M, i, j));  	
	
	/* Right upper corner */
	for(i=1; i <= Row-1; i++)
                for(j=Col+1; j <= MCols(M); j++)
                        SetMatrixElement(C, i, j-1, GetMatrixElement(M, i, j));

	/* Left down corner */
	for(i=Row+1; i <= MRows(M); i++)
                for(j=1; j <= Col-1; j++)
                        SetMatrixElement(C, i-1, j, GetMatrixElement(M, i, j));

	/* Right down corner */
	for(i=Row+1; i <= MRows(M); i++)
                for(j=Col+1; j <= MCols(M); j++)
                        SetMatrixElement(C, i-1, j-1, GetMatrixElement(M, i, j));

	return C;		
} /* MatrixExcept */

double Determinant(MATRIX M) {
/* ---------------------------------------------------------------------------------------------*/
/*                                                                                              */
/*	Matrix determinant.									*/
/*                                                                                              */
/* -------------------------------------------------------------------------------------------- */

	int 	i;
	MATRIX	foo;
	double 	sum = 0;
	
	if ( (MRows(M)==1) && (MCols(M)==1) )
		return GetMatrixElement(M, 1, 1);
	else {
		for(i=1; i <= MCols(M); i++) {
			foo = MatrixExcept(M, 1, i);
			sum += pow(-1,i+1) * GetMatrixElement(M,1,i) 
					   * Determinant(foo);
			DeleteMatrix(foo);
		}
		return sum;
	}
} /* Determinant */

void CopyMatrix(MATRIX From, MATRIX To) {
/* ---------------------------------------------------------------------------------------------*/
/*                                                                                              */
/*	Copies a matrix From to matrix To.							*/
/*	Remember to create destination matrix !							*/
/*                                                                                              */
/* -------------------------------------------------------------------------------------------- */

	int i,j;
	
	if ( (MRows(From) != MRows(To)) || (MCols(From) != MCols(To)) )
		SegFault("ERROR [CopyMatrix]: Error in Matrix copying; size mismatch!");
	
	for(i=1; i <= MRows(From); i++)
		for(j=1; j <= MCols(From); j++)
			SetMatrixElement(To,i,j,GetMatrixElement(From,i,j));
	
} /* CopyMatrix */

static void SwapRows(MATRIX A, int row1, int row2) {
/* ---------------------------------------------------------------------------------------------*/
/*                                                                                              */
/*	Swaps two rows in a matrix.								*/
/*	(A static help function needed in the calculation of inverse matrix).			*/
/*                                                                                              */
/* -------------------------------------------------------------------------------------------- */

    double   tmp; 
    int     i; 

    for(i=1; i <= MCols(A); i++) { 
        tmp = GetMatrixElement(A, row1, i);
	SetMatrixElement(A, row1, i, GetMatrixElement(A, row2, i)); 
	SetMatrixElement(A, row2, i, tmp);
    } 
} /* SwapRows */

MATRIX MatrixInverse(MATRIX Result, MATRIX A) { 
/* ---------------------------------------------------------------------------------------------*/
/*                                                                                              */
/*	Computes an inverse of matrix A.							*/
/*	A(A^-1) = (A^-1)A = I = Identity matrix							*/
/*                                                                                              */
/*	We've got this (algorithm!) from:							*/
/*		        http://www.3dica.org/frenzy/matrix.htm					*/
/*			(23.8.2001)								*/
/*												*/
/* -------------------------------------------------------------------------------------------- */

    int      i, j, k, MaxRow; 
    double    MaxValue, Value; 
    MATRIX   M, I;

    I = NewMatrix(MRows(A), MRows(A));
    IdentityMatrix(I); 

    M = NewMatrix(MRows(A), MCols(A));
    CopyMatrix(A, M);

    MaxValue = GetMatrixElement(M, 1, 1); 
    MaxRow = 1; 
 
    for(i=1; i <= MRows(M); i++) { 

        /* Find row with largest value at the diagonal. */
        MaxValue = GetMatrixElement(M, i, i); 
        MaxRow = i; 
        for(j=i+1; j <= MRows(M); j++) { 
            Value = GetMatrixElement(M, j, i); 
            if(fabs(Value) > fabs(MaxValue)) { 
                MaxValue = Value; 
                MaxRow = j; 
            } 
        } 

        /* Swap the row with largest value with current row. */
        if (MaxRow != i) { 
            SwapRows(M, i, MaxRow); 
            SwapRows(I, i, MaxRow); 
        } 

        /* Divide the entire current row with MaxValue to get a 1 on the diagonal. */
        for(k=1; k <= MCols(M); k++) { 
            	SetMatrixElement(M, i, k, GetMatrixElement(M, i, k)/MaxValue);
		SetMatrixElement(I, i, k, GetMatrixElement(I, i, k)/MaxValue);
        } 

        /* Subtract current row from all other rows so their values before the diagonal go zero. */
        for(j=i+1; j <= MRows(M); j++) { 
            Value = GetMatrixElement(M, j, i); 
            for(k=1; k <= MCols(M); k++) { 
                SetMatrixElement(M, j, k, GetMatrixElement(M, j, k) - 
				GetMatrixElement(M, i, k) * Value); 
		SetMatrixElement(I, j, k, GetMatrixElement(I, j, k) -
                                GetMatrixElement(I, i, k) * Value);
            } 
        } 

    } 

    /* Finally substract values so that the original matrix becomes identity */
    for(i=MRows(M); i >= 1; i--) { 
        for(j=i-1; j >= 1; j--) { 
            Value = GetMatrixElement(M, j, i); 
            for(k=1; k <= MCols(M); k++) { 
		SetMatrixElement(M, j, k, GetMatrixElement(M, j, k) -
                		GetMatrixElement(M, i, k) * Value);
		SetMatrixElement(I, j, k, GetMatrixElement(I, j, k) -
                                GetMatrixElement(I, i, k) * Value);
            } 
        } 
    } 
    
    DeleteMatrix(M);
    CopyMatrix(I, Result);
    DeleteMatrix(I);

    return Result;

} /* MatrixInverse */

double Trace(MATRIX A) {
/* ---------------------------------------------------------------------------------------------*/
/*                                                                                              */
/*	Trace of matrix, i.e. sum of diagonal elements.						*/
/*                                                                                              */
/* -------------------------------------------------------------------------------------------- */

        int     i;
        double   sum = 0;

        for(i=1; i <= MRows(A); i++)
                sum += GetMatrixElement(A, i, i);

        return sum;

} /* Trace */

double VectorDistance(MATRIX x, MATRIX y, int p) {

	int 	i;
	double 	D = 0;

	if (!IsScalar(x) && (IsRowVector(x) || IsRowVector(y)) )
		SegFault("ERROR [VectorDistance] : Use column vectors, please!");
	if ( MRows(x) != MRows(y) )
		SegFault("ERROR [VectorDistance] : dimension mismatch!");
	if ( p <= 0)
		SegFault("ERROR [VectorDistance] : p should be > 0");
	
	for(i=1; i <= MRows(x); i++)
		D += pow(fabs(GetMatrixElement(x, i, 1) - GetMatrixElement(y, i, 1)), p);
	
	return pow(D,1.0/p);	

}

int Same(MATRIX A, MATRIX B) {

	int 	i,j;
	double	Err;
	double	Tolerance = 1e-6;

	if ( (MRows(A) != MRows(B)) || (MCols(A) != MCols(B)) )
		return 0;
	if (A == B)
		return 1;
	for(i=1; i <= MRows(A); i++) 
		for(j=1; j <= MCols(A); j++) {
		   Err = fabs( (GetMatrixElement(A,i,j) - GetMatrixElement(B,i,j))
			/ MAX(GetMatrixElement(A,i,j),GetMatrixElement(B,i,j)) );
		   if ( Err >= Tolerance )
			return 0;
		}

	return 1;

}

double Avg(MATRIX Vector) {
/* Average value of Vector */

	int 	i, NumElem;
	double 	sum = 0;

	NumElem = -1; /* Compiler doesn't warn 
			 about uninitialization :) */

	if (IsColVector(Vector)) {
	   NumElem = MRows(Vector);
	   for(i=1; i <= NumElem; i++)
	   	sum += GetMatrixElement(Vector,i,1);
	}
	else if (IsRowVector(Vector)) {
           NumElem = MCols(Vector);
	   for(i=1; i <= NumElem; i++)
		sum += GetMatrixElement(Vector,1,i);
	}
	else
	   SegFault("ERROR[Avg]: You should give a vector as input.");

	return sum/NumElem;
}

double Var(MATRIX Vector, int UnbiasedEstimate) {
/* Variance of the random variables given in the Vector.

   Vector is type (N x 1).
   UnbiasedEstimate = 1:
	Normalize by N-1
   Else
	Normalize by N.
*/

	int     i, NumElem;
	double	Mean, SqSum = 0;

        NumElem = -1; /* Compiler doesn't warn
                         about uninitialization :) */
   
        if (IsColVector(Vector))
           NumElem = MRows(Vector);
        else if (IsRowVector(Vector))
           NumElem = MCols(Vector);
	else
           SegFault("ERROR[Var]: You should give a vector as input.");	

	Mean = Avg(Vector);
	for(i=1; i <= NumElem; i++) {
		if (IsColVector(Vector))
			SqSum += pow(GetMatrixElement(Vector,i,1)
					- Mean, 2);
		else
        		SqSum += pow(GetMatrixElement(Vector,1,i)
					- Mean, 2);
	}
        
	if (UnbiasedEstimate)
        	return SqSum/(NumElem-1);
	else
		return SqSum/NumElem;		

}

double InnerProduct(MATRIX Vec1, MATRIX Vec2) {
	
	MATRIX 	InProd;
	double	Result;

	if (!IsRowVector(Vec1) && !IsColVector(Vec2)) 
		SegFault("ERROR [InnerProduct]: inputs should \
			be : Vec1 = (1 x d), Vec2 = (d x 1).");

	InProd = NewMatrix(1,1);	

	MatrixProduct(InProd, Vec1, Vec2);
	Result = GetMatrixElement(InProd,1,1);

	DeleteMatrix(InProd);

	return Result;
}

double Norm(MATRIX Vector) {

	double	Result;
	MATRIX  Foo;

	if (!IsColVector(Vector))
		SegFault("ERROR [Norm] : give a col vector.\n");

	Foo = NewMatrix(MRows(Vector),1);

	CopyMatrix(Vector, Foo);
	Result = pow( InnerProduct(Transpose(Foo), Vector), 0.5);

	DeleteMatrix(Foo);
	
	return Result;
}

MATRIX  OneMatrix(MATRIX M) {
 
        int     i, j;
        
        for(i=1; i <= MRows(M); i++)
                for(j=1; j <= MCols(M); j++)
                        SetMatrixElement(M, i, j, 1);
        
        return M;

}

double ProdOfElements(MATRIX M) {

	int 	i, j;
	double 	Prod;

	Prod = 1;
	for(i=1; i <= MRows(M); i++)
	   for(j=1; j <= MCols(M); j++)
	      Prod *= GetMatrixElement(M,i,j);
	
	return Prod;   
}


double SumOfElements(MATRIX M) {

        int     i, j;
        double   Sum;
 
        Sum = 0;
        for(i=1; i <= MRows(M); i++)   
           for(j=1; j <= MCols(M); j++)   
              Sum += GetMatrixElement(M,i,j);   
                 
        return Sum;
}

void SquareElements(MATRIX M) {
    double* Loc = M->Data + (M->Rows * M->Cols);
    while (Loc-- > M->Data) *Loc = *Loc * *Loc;
}

double  VectorSimilarity( MATRIX Vec1, MATRIX Vec2,
                        VECTOR_SIMILARITY_TYPE SimType,
			double Alpha ) {
   double dist, Val = 0;

   if ( MCols(Vec1) > 1 || MCols(Vec2) > 1 )
      SegFault("ERROR [VectorSimilarity]: Column vector expected.");
   else if ( MRows(Vec1) != MRows(Vec2) )
      SegFault("ERROR [VectorSimilarity]: Dimension mismatch of vectors.");
   else if (Alpha <= 0)
      SegFault("ERROR [VectorSimilarity]: Alpha must be positive.");

   switch(SimType) {
      case smooth_Gauss:
	   dist = VectorDistance( Vec1, Vec2, 2); // Euclidean distance
	
	   /* Normalize squared distance by dimensionality */
	   dist = ( dist * dist ) / (1.0 * MRows(Vec1));

           Val = (1.0/Alpha) * ( (exp(-0.5*( pow((dist/1.0*Alpha),2) ))) - 1) + 1;
	   break;
      case Lp_similarity:
	   dist = VectorDistance( Vec1, Vec2, 2); // Euclidean distance
	
	   dist = dist * dist; // Get rid of sqrt(.)
	   
	   Val = pow( (1 + dist), -2.0 );
	   break;
 
   }

   return Val;

}

void NormalizeSumToUnity( MATRIX M ) {
/* After normalization, sum of elements of M is 1. */
   
   int   i, j;
   double Sum = SumOfElements(M);

   for(i=1; i <= MRows(M); i++)
      for(j=1; j <= MCols(M); j++)
         SetMatrixElement( M, i, j, 
	    GetMatrixElement( M, i, j )/Sum
	 );
}
