/* --------------------------------------------------------------------*/
/*  mel-functions.c       Teemu Kilpelinen 22.5.2001		       */
/*  some modifications by Ville Hautamki, <villeh@cs.joensuu.fi, 2002 */
/*                                                                     */
/*  This module includes functions for counting of mel.		       */
/*	Version Number: 0.01					       */
/* --------------------------------------------------------------------*/

#include <float.h>
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "speakerid.h"
#include "util.h"
#include "mel-functions.h"

/* --------------------------------------------------------------- */
/*   
    Functioning:    Returns pointer to array filled by mel-coeffs.	
	Inputs:		    int begin	Index where from to begin 
					int center	Index of center point of filter
					int end	    Index where to stop
	Outputs:		-
	Return value:	Pointer to array filled by mel-coeffs.
	History:		Teemu Kilpelinen 22.05.2001 
	Version:		1.0 (original)
*/	
static double *FillCoeffTable(int begin, int center, int end)
{
  int i;
  double *coeffs;
	
  //  printf("%d %d Size: %d\n", end, begin,(end - begin + 1));
  coeffs = (double *) allocate(sizeof(double)*(end - begin + 1), NO_INIT);
		
  for (i = 1; i <= center - begin; i++)
    coeffs[i - 1] = i/(double)(center - begin);
	
  for (i = 0; i <= end - center; i++)
    coeffs[center - begin + i] = (end - center - i)/(double)(end - center + 1);

  return coeffs;
} 

		
/* --------------------------------------------------------------- */
/*   
    Functioning:    Returns array filled by mel-coeffs.	
	Inputs:		    int count_of_filters 
					int length (pow of two)  
					int sampling_rate
	Outputs:		-
	Return value:	Filter bank.
	History:		Teemu Kilpelinen ja Ismo Krkkinen 22.05.2001 
	Version:		1.0 (original)
*/	

Filterbank *CreateFilterBank(int count_of_filters, int length, 
			     int sampling_rate)
{
  int k, prev_center = 0, next_center, curr_center;
  double mel_max, mel_freq_Hz, fmel;
  Filterbank* filter_bank;

  //  printf("srate: %d\n", sampling_rate);

  mel_max = Hz2Mel(sampling_rate/2.0f);   

  filter_bank = allocate(sizeof(Filterbank), NO_INIT);

  filter_bank->num_mel_filters = count_of_filters;
  filter_bank->filters = allocate(sizeof(Filter)*count_of_filters, NO_INIT);

  fmel = 1/((double)count_of_filters) * mel_max;
  mel_freq_Hz = (double)(700 * (pow(10,(fmel/2595.0)) - 1));
  curr_center = Hz2Index(mel_freq_Hz, sampling_rate, length);

  for (k = 1; k <= count_of_filters; k++) {
    fmel = (k + 1)/((double)count_of_filters) * mel_max;
    mel_freq_Hz = (double)(700 * (pow(10,(fmel/2595.0)) - 1));
		
    if (k != count_of_filters)
      next_center = Hz2Index(mel_freq_Hz, sampling_rate, length);
    else
      next_center = curr_center;

    //    printf("%d %d\n", curr_center, next_center-curr_center);

    filter_bank->filters[k - 1].filter_begin = prev_center;
    filter_bank->filters[k - 1].filter_end   = next_center;
    filter_bank->filters[k - 1].coeffs       = 
      FillCoeffTable(prev_center, curr_center, next_center);

    /* Fill coefficient array */
    prev_center = curr_center;
    curr_center = next_center;
  }
	
  filter_bank->filters[count_of_filters - 1].filter_end = 
    (int)(length/2.0 - 1);

  return filter_bank;
} 


/* --------------------------------------------------------------- */
/*	
    Functioning:    Returns array filled by mel-coeffs.	
	Inputs:		    Filterbank* filter_bank	Filter bank.
					int filter_index			Index in filter bank.
	Outputs:		-
	Return value:	One filter from filter_bank.
	History:		Teemu Kilpelinen 22.05.2001 
	Version:		1.0 (original)
*/
Filter *GetFilter(Filterbank *filter_bank, int filter_index)
{
  return &(filter_bank->filters[filter_index - 1]);
}


/* 
	Functioning:    Filters frame with one filter.	
	Inputs:		    Filter* filter		Pointer to one filter.
					DOUBLE_ARRAY frame   Frame (array where the piece 
									    of signal have been copied).  
	Outputs:		
	Return value:	-
	History:		Teemu Kilpelinen 22.05.2001 
	Version:		1.0 (original)
*/
double FilterOutput(Filter *filter, double *frame)
{
  int k;
  double sum = 0;

  for (k = filter->filter_begin; k <= filter->filter_end; k++) {
    sum = sum + (frame[k] * filter->coeffs[k - filter->filter_begin]);
  }

  /* the average of all result values. */
  return sum/(filter->filter_end - filter->filter_begin + 1);
}


/* --------------------------------------------------------------- */
/* 
    Functioning:    Returns the number of Mel-Filters 	
	Inputs:		    int sampling_rate  Sampling rate of input signal.	    
	Outputs:		-
	Return value:	The number of Mel-Filters
	History:		Teemu Kilpelinen 01.03.2001
	Version:		1.0 (original)
*/
int NumberOfMelFilters(int sampling_rate)
{
  return (int)ceil((double)(3 * log(sampling_rate)));
} 


/* --------------------------------------------------------------- */
/*
	Functioning:    Computes the log-filtered spectrum of spectrum (fft's result)
	Inputs:			FRAME_ARRAY frame
					DOUBLE_ARRAY log_filtered_spectrum
					Filterbank* filter_bank
					USER_PARAMETERS user_params		User parameters
    Outputs:		An array which includes the values of log-filtered frame.

	Return value:	- 
	History:		Teemu Kilpelinen 11.04.2001
	Version:		1.0 (original)

	Reference:		Discrete-Time Processing of Speech Signals, 
					John R. Deller Jr., John H.L. Hansen, 
					John G. Proakis, 2000.  
*/
void ComputeLogFilteredSpectrum(double *frame, double *log_filtered_spectrum, 
				Filterbank *filter_bank, Parameters *parap)
{ 
  int i;  

  /* Absolute magnitude domain */
  if (parap->featptr->absolute_magnitude == 1) {
    for (i = 1; i <= filter_bank->num_mel_filters; i++) {
      log_filtered_spectrum[i - 1] = 
	(double)log10(FilterOutput(GetFilter(filter_bank, i), frame)+FLT_EPSILON);
    }
	
  } else /* Power domain */
    for (i = 1; i <= filter_bank->num_mel_filters; i++){
      log_filtered_spectrum[i - 1] = 
	(double)log10(pow(FilterOutput(GetFilter(filter_bank, i), frame), 2)+FLT_EPSILON);
    }
}

void free_filterbank(Filterbank *filterbank)
{
  int i;
  Filter *temp;

  //  printf("num: %d \n", filterbank->num_mel_filters);

  for (i = 0; i < filterbank->num_mel_filters; i++) {
    temp = GetFilter(filterbank, i + 1);
    free(temp->coeffs);
  }

  free(filterbank->filters);
  free(filterbank);
}

/* --------------------------------------------------------------- */
/*  
	Functioning:	Changes Hz to mel-scale.
	Inputs:			double frequency        In Hz.
					int sampling_rate      In Hz.
				    int frame_length       frame length in samples (power of two).	
					(WITHOUT decimation by 2 == window with added zeroes).	
	Outputs:		
	Return value:	index of the frequency in the FFT table.  
	History:		Teemu Kilpelinen      20.12.2000, 17.3.2001
	Version:		1.0 (original)

	Reference:		Discrete-Time Signal Processing of Speech Signals, 1999. 
					IEEE Press.
*/	  
int Hz2Index(double frequency, int sampling_rate, int frame_length)
{
  double Nyquist_freq; /* Nyquist-frequency of the signal */
  double freq_step;    /* Frequency step */
  int rounded_index;  /* round of nearest index */

  // printf("Hz2Index: %f %d %d\n",frequency, sampling_rate, frame_length);

  Nyquist_freq = sampling_rate/2.0f;  
	
  freq_step = Nyquist_freq/(frame_length/2.0f); 
		
  /* Round of nearest index */	
  rounded_index = (int)(frequency / freq_step);           
    
  /* if rounded_index is over frame_length, returns frame_length */
  if (rounded_index < frame_length/2)
    return rounded_index;
  else 
    return rounded_index; 
}


/* --------------------------------------------------------------- */
/*
	Functioning:	f(mel) = 2595 * log10(1+ f/700)
	Inputs:			double Hz_frequency     Frequency in hertz.  
	
	Outputs:		
	Return value:	 
	History:		Teemu Kilpelinen and Tomi Kinnunen		22.01.2001
	Version:		1.0 (original)

	Reference: Discrete-Time Signal Processing of Speech Signals, 1999. 
			   IEEE Press.		
*/
double Hz2Mel(double Hz_frequency)
{
  return 2595.0f * (double)log10(1 + Hz_frequency/700.0);
} 

