#include <stdio.h>
#include <stdlib.h>

#include "pattern.h"
#include "misc.h"
#include "matrix.h"
#include "fvec.h"
#include "textfile.h"

PATTERN NewPattern(int Dimensionality, int NumVectors) {

	PATTERN X = malloc(sizeof(PATTERN_TYPE));

	if ((Dimensionality > 0) && (NumVectors >= 0)) {

		X->NumberOfVectors = NumVectors;
		X->Dimensionality = Dimensionality;
		X->Data = NewMatrix(Dimensionality, NumVectors);
	}

	else SegFault("Failed to create new pattern (Pattern size or dimensionality <= 0)");

	return X;

}

void PrintPattern(PATTERN X, FILE* f) {

	MATRIX Foo;
	
	Foo = NewMatrix(MRows(X->Data),MCols(X->Data));
	CopyMatrix(X->Data, Foo); 

	PrintMatrix(Transpose(Foo), f);
	
	DeleteMatrix(Foo);

}

void PrintPatternSet(FILE* f, PATTERN_SET PS) {
	
   int i;

   for(i=1; i <= PSNumPatterns(PS); i++)
      PrintPattern( GetPattern(PS, i), f );
}

void DeletePattern(PATTERN X) {

	DeleteMatrix(X->Data);
	free(X);
}

PATTERN LoadPattern(char* Filename) {

	FILE*	File;
	int	Dimensionality, VectorCount, i, j;
	double**	Data;
	PATTERN X;

	File = fopen(Filename, "rt");
	if (File != NULL) {
		switch(fvReadSet(&Data, &VectorCount, &Dimensionality, File)) 
		{
		   case 0: // Error in file (such as letters, ...)
		      fclose(File);
		      fprintf(stderr, "ERROR reading file %s\n", Filename);
		      return NULL;
		      break;
                   case -1: // Error in file, lines were removed
		      fprintf(stderr, "WARNING: file %s has incomplete lines.\n", Filename);
		      break;
		}
		fclose(File);
	}
	else {
		fprintf(stderr, "ERROR opening file %s\n", Filename);
		return NULL;
	}

	X = NewPattern(Dimensionality, VectorCount);
	for(i=1; i <= VectorCount; i++)
		for(j=1; j <= Dimensionality; j++)
			SetMatrixElement(X->Data,j,i,Data[i-1][j-1]);
	
	fvDeleteSet(Data, VectorCount);

	return X;	
}

MATRIX GetVector(MATRIX TargetVector, PATTERN X, int Index) {

	if ((Index <= 0) || (Index > PNumVectors(X))) 
		SegFault("Vector index out of pattern range.");
	else if (!(IsColVector(TargetVector)))
		SegFault("ERROR [GetVector]: Target vector must be column vec!");	

	GetColumn(TargetVector, X->Data, Index);
	
	return TargetVector;

}

MATRIX MeanVector(MATRIX TargetVector, PATTERN X) {
/* Compute mean vector of the pattern */

	int i;
	MATRIX y;
	
	if (MRows(TargetVector) != PDimensionality(X))
		SegFault("ERROR [MeanVector] : dimension mismatch.");

	y = NewMatrix(PDimensionality(X),1);
	ZeroMatrix(TargetVector);
	
	for(i=1; i <= PNumVectors(X); i++) {
		GetVector(y, X, i);
		MatrixSum(TargetVector, TargetVector, y);	
	}
	MultiplyByScalar(TargetVector, TargetVector, 1.0/PNumVectors(X));
	
	DeleteMatrix(y);
	
	return TargetVector;
}

PATTERN SetVector(PATTERN X, MATRIX Vector, int Index) {

	if ((Index <= 0) || (Index > PNumVectors(X))) 
                SegFault("Vector index out of pattern range.");
	else if (!IsColVector(Vector))
		SegFault("Vector must be a column vector in assignment to pattern!");

	SetColumn(X->Data, Vector, Index);

	return X;	

}

PATTERN InsertNewVector(PATTERN X, MATRIX Vector) {

	InsertNewVectorToMatrix(X->Data, Vector,1);
	X->NumberOfVectors++;

	return X;
}

PATTERN PickRandomVectors(PATTERN Target, PATTERN Source) {

	int To, From, k;
	if (PNumVectors(Source) < PNumVectors(Target))
		SegFault("ERROR [PickRandomVectors]: Source pattern has
				too few vectors !");

	for (To = 1; To <= PNumVectors(Target); ++To) {
	    From = 1 + PNumVectors(Source) * (rand() / (double)RAND_MAX);
	    for (k = 1; k < PDimensionality(Target); ++k)
		SetMatrixElement(Target->Data, k, To,
		    GetMatrixElement(Source->Data, k, From));
	}

	return Target;

}

void CopyVectors( PATTERN Target, PATTERN Source, 
		  int StartPos, int EndPos) {
/* 
   Copy vectors from pattern to another pattern.
   
   Target   = Target pattern,
   Source   = Source pattern,
   StartPos = Index where to start copying
   EndPos   = Index where to stop  copying

*/

   int 	   i, NumVectors, TargetIndex;
   MATRIX  Vector;

   NumVectors = EndPos - StartPos + 1;

   if ( (StartPos < 1) || (StartPos > PNumVectors(Source)) )
      SegFault("ERROR [CopyVectors] : StartPos out of range.");
   else if ( (EndPos < 1) || (EndPos > PNumVectors(Source)) )
      SegFault("ERROR [CopyVectors] : EndPos out of range.");
   else if ( StartPos > EndPos )
      SegFault("ERROR [CopyVectors] : StartPos > EndPos.");
   else if (PNumVectors(Target) < NumVectors)
      SegFault("ERROR [CopyVectors] : Target pattern has too few vectors.");
   else if (PDimensionality(Target) != PDimensionality(Source))
      SegFault("ERROR [CopyVectors] : Dimension mismatch of pattern.");   

   Vector = NewMatrix( PDimensionality(Target), 1 );
   TargetIndex = 1;
   for(i=StartPos; i <= EndPos; i++) {
      GetVector(Vector, Source, i);
      SetVector(Target, Vector, TargetIndex);
      TargetIndex++;
   } 

   DeleteMatrix(Vector);

}




MATRIX CovMatrix(MATRIX TargetMatrix, PATTERN X) {

	MATRIX	MeanVec, Vector, Foo, Foo2, Foo3;
	int	i;

	if (!IsSquare(TargetMatrix))
		SegFault("ERROR [CovMatrix] : Non-square target matrix.");
	if (MRows(TargetMatrix) != PDimensionality(X))
                SegFault("ERROR [CovMatrix] : dimension mismatch.");


	MeanVec = NewMatrix(PDimensionality(X), 1);
	Vector  = NewMatrix(PDimensionality(X), 1);
	ZeroMatrix(TargetMatrix);
	Foo = NewMatrix(PDimensionality(X), 1);
	Foo2 = NewMatrix(PDimensionality(X), PDimensionality(X));
	Foo3 = NewMatrix(PDimensionality(X), 1);

	MeanVector(MeanVec, X);
	for(i=1; i <= PNumVectors(X); i++) {
		GetVector(Vector, X, i);
		MatrixSubstract(Foo, Vector, MeanVec);
		CopyMatrix(Foo, Foo3);
		MatrixProduct(Foo2, Foo, Transpose(Foo3));
		MatrixSum(TargetMatrix, TargetMatrix, Foo2);
		Transpose(Foo3);
	}	
	MultiplyByScalar(TargetMatrix, TargetMatrix, 1.0/PNumVectors(X));	

	DeleteMatrix(MeanVec);
	DeleteMatrix(Vector);
	DeleteMatrix(Foo);
	DeleteMatrix(Foo2);
	DeleteMatrix(Foo3);

	return TargetMatrix;
}

MATRIX Variance(MATRIX TargetVector, PATTERN X) {

	MATRIX	MeanVec, Vector;
	int	i, j;
	double	sum, MeanElem;

	if (MRows(TargetVector) != PDimensionality(X))
                SegFault("ERROR [Variance] : dimension mismatch.");

	MeanVec = NewMatrix(PDimensionality(X), 1);
	MeanVector(MeanVec, X);
	Vector  = NewMatrix(PDimensionality(X), 1);	

	for(i=1; i <= PDimensionality(X); i++) {
		sum = 0;
		MeanElem = GetMatrixElement(MeanVec,i,1);
		for(j=1; j <= PNumVectors(X); j++) {
			GetVector(Vector, X, j);
			sum += pow(GetMatrixElement(Vector,i,1)-MeanElem, 2);
		}
		sum *= 1.0/PNumVectors(X);
		SetMatrixElement(TargetVector,i,1,sum);
	}
	
	DeleteMatrix(MeanVec);
	DeleteMatrix(Vector);

	return TargetVector;

}

MATRIX StandardDev(MATRIX TargetVector, PATTERN X) {

	MATRIX 	Var;
	int	i;

	if (MRows(TargetVector) != PDimensionality(X))
                SegFault("ERROR [StandardDeviation] : dimension mismatch.");

	Var = NewMatrix(PDimensionality(X), 1);
	Variance(Var, X); 
	
	for(i=1; i <= PDimensionality(X); i++)
		SetMatrixElement(TargetVector,i,1,
			pow(GetMatrixElement(Var,i,1),0.5) );
	
	DeleteMatrix(Var);

	return TargetVector;

}

double ComputeMeanDistVariance(PATTERN X) {
/*	
	Q: What the heck is this function? 8-)
	A: The algorithm:
		1. Compute the mean vector of X,
		2. Compute distances to the mean,
		3. Return variance of the distances.

	Note also:
		Unbiased version of the variance is used:
		Var = (1/(N-1)) * SUM { (xi - mean)^2 }
*/

	int 	i;
	double	AvgDist, Result;
	MATRIX 	Vector, Centroid;

	if (PNumVectors(X) < 2)
		SegFault("ERROR[ComputeMeanDistVariance]:\
			pattern must have at least two vectors!");

	Vector = NewMatrix(PDimensionality(X),1);
	Centroid = NewMatrix(PDimensionality(X),1);

	MeanVector(Centroid, X);
	AvgDist = 0;
	for(i=1; i <= PNumVectors(X); i++) {
	   GetVector(Vector, X, i);
	   AvgDist += VectorDistance(Vector,Centroid,2);
	}
	AvgDist /= PNumVectors(X);

	Result = 0;
	for(i=1; i <= PNumVectors(X); i++) {
           GetVector(Vector, X, i);
           Result += pow( VectorDistance(Vector,Centroid,2)
			   - AvgDist, 2 );
        }

	DeleteMatrix(Vector);
	DeleteMatrix(Centroid);

	return Result/(PNumVectors(X)-1);
}  

int FindNearestVector(MATRIX Vector, PATTERN P) {       
/* Index of the nearest vector to "Vector" in pattern "P" */

        int             i, NearestIndex;
        double           Foo, Minimum;
        MATRIX          y;

        if (!IsColVector(Vector))
                SegFault("ERROR [FindNearestVector]: Input is not a column
+vector\n");
        if (PDimensionality(P) != MRows(Vector))
                SegFault("ERROR [FindNearestVector]: Unequal
+dimensionalities.\n");

        y = NewMatrix(PDimensionality(P),1);

        GetVector(y,P,1);
        NearestIndex = 1;
        Minimum = VectorDistance(Vector, y, 2); /* Euclidean norm */
        for (i=2; i <= PNumVectors(P); i++) {
                GetVector(y, P, i);
                Foo = VectorDistance(Vector, y, 2);
                if (Foo < Minimum) {
                        Minimum = Foo;
                        NearestIndex = i;
                }
        }

        DeleteMatrix(y);

        return NearestIndex;
}


double DistanceToNearestVector(MATRIX Vector, PATTERN P) {

	MATRIX Foo;	
	double  Dist;	

	if (!IsColVector(Vector))
                SegFault("ERROR [DistanceToNearestVector]: Input is not a column vector\n");
        if (PDimensionality(P) != MRows(Vector))
                SegFault("ERROR [DistanceToNearestVector]: Unequal dimensionalities.\n");

	Foo = NewMatrix(PDimensionality(P),1);

	GetVector(Foo, P, FindNearestVector(Vector,P));

	Dist = VectorDistance(Vector, Foo, 2); 
	
 	DeleteMatrix(Foo);
	return Dist;

}

double MeanSquaredError(PATTERN X, PATTERN C) {
	
	int	i;
	MATRIX	FooVec;
	double 	Sum = 0;

	if (PDimensionality(X) != PDimensionality(C))
		SegFault("ERROR [MeanSquaredError] : Unequal dimensionalities.\n");
	
	FooVec = NewMatrix(PDimensionality(X),1);

	for(i=1; i <= PNumVectors(X); i++) {
		GetVector(FooVec, X, i);
		Sum += pow( DistanceToNearestVector(FooVec, C), 2);
	}	

	DeleteMatrix(FooVec);

	return Sum/PNumVectors(X);
	
}

void ZeroMeanUnitVarianceNormalize(PATTERN X) {
/* For each dimension: remove mean and divide by
   std --> E(Xi) = 0 and Var(Xi) = 1 */

	MATRIX 	Mean;
	MATRIX	Stdev;
	MATRIX	Vec;
	int	i, j;

	Mean  = NewMatrix(PDimensionality(X),1);
	Stdev = NewMatrix(PDimensionality(X),1);	
	Vec   = NewMatrix(PDimensionality(X),1);

	Stdev = StandardDev(Stdev, X);
	Mean  = MeanVector(Mean, X);

	for(i=1; i <= PNumVectors(X); i++) {
		GetVector(Vec, X, i);
		for(j=1; j <= PDimensionality(X); j++)
		   SetMatrixElement( Vec,j,1,
		      (GetMatrixElement(Vec,j,1) - 
		       GetMatrixElement(Mean,j,1))/
		      (GetMatrixElement(Stdev,j,1))
		   );		
		SetVector(X, Vec, i);
	}

	DeleteMatrix(Mean);
	DeleteMatrix(Stdev);
	DeleteMatrix(Vec);

}

PATTERN TranslatePattern(PATTERN X, MATRIX Vector) {
	
	int 	i;
	MATRIX 	Vec, TransVec;
	
	if (PDimensionality(X) != MRows(Vector))
		SegFault("ERROR [TranslatePattern]: dimension mismatch!");
		
	Vec = NewMatrix(PDimensionality(X),1);
	TransVec = NewMatrix(PDimensionality(X),1);
	for(i=1; i <= PNumVectors(X); i++) {
		GetVector(Vec, X, i);
		MatrixSum(TransVec, Vec, Vector);
		SetVector(X, TransVec, i);
	}

	DeleteMatrix(TransVec);
	DeleteMatrix(Vec);

	return X;

}

#if 0
MATRIX ScatterMatrix(MATRIX TargetMatrix, PATTERN X) {

	int 	i;
	MATRIX	MeanVec, Foo1, Foo2, Vector;

	MeanVec = NewMatrix(PDimensionality(X),1);
	Foo1 = NewMatrix(PDimensionality(X),1);
	Foo2 = NewMatrix(PDimensionality(X),1);
	Vector = NewMatrix(PDimensionality(X),1);

	/* Make TargetMatrix zero */
	ZeroMatrix(TargetMatrix);

	/* Compute class mean */
	MeanVec = MeanVector(MeanVec, X);

	for(i=1; i <= PNumVectors(X); i++) {
		GetVector(Foo1, X, i);
		Foo2 = MatrixSubstract(Foo2, Foo1, MeanVec);
		TargetMatrix = MatrixSum(TargetMatrix,
			Foo2, Transpose(Foo2);	   	   	
		Transpose(Foo2);
	}	

	DeleteMatrix(Foo1);
	DeleteMatrix(Foo2):
	DeleteMatrix(MeanVec);
	DeleteMatrix(Vector);

	return TargetMatrix;

}
#endif

int FindMostSimilarVector(MATRIX Vector, PATTERN P) {
/* See Duda & Hart 2001, p. 541:
	
   Vector similarity = s(x1,x2) = (x1'x2)/(||x1||*||x2||)	

*/
	int	i, MaxSimInd;
	double 	Sim, MaxSim;
	MATRIX	Vec, Vec2; 

	Vec = NewMatrix(PDimensionality(P),1);
	Vec2 = NewMatrix(PDimensionality(P),1);
	
	CopyMatrix(Vector, Vec2);
	Transpose(Vec2);

	MaxSimInd = 1; 
	MaxSim = -1;
	for(i=1; i <= PNumVectors(P); i++) {
		GetVector(Vec, P, i);
		Sim = InnerProduct(Vec2, Vec)/
		      (Norm(Vec) * Norm(Vector));
		if (Sim > MaxSim) {
			MaxSim = Sim;
			MaxSimInd = i;
		}
	}

	DeleteMatrix(Vec);
	DeleteMatrix(Vec2);

	return MaxSimInd;
}

double BhattacharyyaDistance_Gauss( MATRIX Cov1,  MATRIX Cov2,
                                   MATRIX Mean1, MATRIX Mean2 ) {
/* ---------------------------------------------------------------------- */
/*                                                                        */
/*      Battacharyaa distance between two Gaussian multivariate           */
/*      density functions N(Mean1, Cov2) and N(Mean2, Cov2).              */
/*									  */
/*	ASSUMEs diagonal covariance matrices !!!!!                        */
/*                                                                        */
/* ---------------------------------------------------------------------- */
        
        double   FooDouble,
                FooDouble2,
                FooDouble3;
        int     i;
        MATRIX  FooMatrix;

        FooMatrix = NewMatrix(MRows(Cov1), 1);

        MatrixSum(FooMatrix, Cov1, Cov2);
        MultiplyByScalar(FooMatrix, FooMatrix, 0.5);
        FooDouble = Determinant(FooMatrix);
                
        FooDouble = 0.5 * log( FooDouble / sqrt( Determinant(Cov1) * Determinant(Cov2) ) );
        
	/* "Matrix inversion" */
        for(i=1; i <= MRows(Cov1); i++)
           SetMatrixElement(FooMatrix,i,1, 1.0/GetMatrixElement(FooMatrix,i,1) );
        
        FooDouble2 = 0;
        for(i=1; i <= MRows(Cov1); i++) {
            FooDouble3 = GetMatrixElement(Mean1, i, 1) -
                        GetMatrixElement(Mean2, i, 1);
            FooDouble3 *= FooDouble3;
            FooDouble3 *= GetMatrixElement(FooMatrix,i,1);
            FooDouble2 += FooDouble3;
        }
           
        return FooDouble + 0.125 * FooDouble2;
        
}

PATTERN_SET NewPatternSet() {

    PATTERN_SET PS = malloc(sizeof(PATTERN_SET_TYPE));
    PS->NumberOfPatterns = 0;
    PS->Data = NULL;
        
    return PS;   

}

void DeletePatternSet(PATTERN_SET PS) {
    
    int i;
    
    for(i=1; i <= PSNumPatterns(PS); i++)
       DeletePattern(PS->Data[i-1]);
    
    free(PS->Data);
    free(PS);
}

PATTERN GetPattern(PATTERN_SET PS, int Index) {

    if ( (Index > 0) && (Index <= PSNumPatterns(PS)) )
       return PS->Data[Index-1];
    else
       SegFault("ERROR [GetPattern]: Pattern index out of range.");
    
    return NULL;
}

void InsertNewPattern(PATTERN P, PATTERN_SET PS) {

   if (PS->NumberOfPatterns == 0) {
      PS->Data = malloc(sizeof(PATTERN));
      PS->NumberOfPatterns = 1;
      PS->Data[0] = P;
   }
   else {
      PS->Data = realloc(PS->Data, sizeof(PATTERN) 
		* (PS->NumberOfPatterns + 1));
      PS->NumberOfPatterns++;
      PS->Data[PS->NumberOfPatterns-1] = P;
   }   

}

void NormalizePatternsZmUv(PATTERN_SET PS) {
/* Zero mean, unit variance normalization for all patterns */

   MATRIX TotalMean, Mean, Std, Vec;
   int	  i, j, k, TotalNumVectors;
   double  Sum, MeanElem, Elem;

   /* Compute mean vector */
   TotalNumVectors = 0;
   Mean = NewMatrix(PDimensionality(GetPattern(PS,1)), 1);
   TotalMean = NewMatrix(PDimensionality(GetPattern(PS,1)), 1);
   ZeroMatrix(TotalMean);
   for(i=1; i <= PSNumPatterns(PS); i++) {
      MeanVector(Mean, GetPattern(PS, i));
      MultiplyByScalar(Mean, Mean, PNumVectors(GetPattern(PS, i)));
      MatrixSum(TotalMean, TotalMean, Mean);
      TotalNumVectors += PNumVectors(GetPattern(PS, i));
   }
   MultiplyByScalar(TotalMean, TotalMean, 1.0/TotalNumVectors);

   /* Compute standard deviation vector */
   /* NOTE: unbiased estimate is used (normalization by N-1) */
   Std = NewMatrix(PDimensionality(GetPattern(PS,1)), 1);
   Vec = NewMatrix(PDimensionality(GetPattern(PS,1)), 1);
   for(k=1; k <= PDimensionality(GetPattern(PS, 1)); k++) {
      Sum = 0;
      MeanElem = GetMatrixElement(TotalMean, k, 1);
      for(i=1; i <= PSNumPatterns(PS); i++)
         for(j=1; j <= PNumVectors(GetPattern(PS, i)); j++) {
             GetVector(Vec, GetPattern(PS, i), j);
             Sum += pow(GetMatrixElement(Vec,k,1)-MeanElem, 2);   	
         }
      Sum *= 1.0/(TotalNumVectors-1);
      SetMatrixElement(Std, k, 1, sqrt(Sum));
   }
   
   /* Do normalization */
   for(i=1; i <= PSNumPatterns(PS); i++) {
      ProgressBar(stdout, i, PSNumPatterns(PS)); 
      for(j=1; j <= PNumVectors(GetPattern(PS, i)); j++) {
         GetVector(Vec, GetPattern(PS, i), j);
	 for(k=1; k <= PDimensionality(GetPattern(PS, 1)); k++) {
            Elem = GetMatrixElement(Vec,k,1);
            Elem -= GetMatrixElement(TotalMean,k,1);
            Elem /= GetMatrixElement(Std,k,1);
	    SetMatrixElement(Vec,k,1,Elem);
         }
         SetVector(GetPattern(PS,i), Vec, j);
      }
   }

   DeleteMatrix(TotalMean);
   DeleteMatrix(Mean);
   DeleteMatrix(Std);
   DeleteMatrix(Vec);

}

PATTERN_SET LoadDataSets( char *FileListName ) {         
/* ---------------------------------------------------------------------- */
/*                                                                        */
/*      Load data sets to memory.                                         */
/*                                                                        */
/* ---------------------------------------------------------------------- */

   char        **Filenames;
   int         i, FileCount;
   PATTERN     P;
   PATTERN_SET PS;

   Filenames = ReadTextFileLines(FileListName, &FileCount);            
   if (FileCount >= 1)
      PS = NewPatternSet();
   else
      return NULL;
   
   fprintf(stdout, "Loading %i data files ...\n", FileCount);         
   for(i=1; i <= FileCount; i++) {
      ProgressBar(stdout, i, FileCount);
      P = LoadPattern(Filenames[i-1]);
      if (P != NULL)
         InsertNewPattern(P, PS);
      else
         return NULL;
   }
   fprintf(stdout, "\n");
         
   return PS;
   
}

double DatasetSimilarity( PATTERN X, PATTERN C, 
                         MATRIX  WeightTable ,
			 int     Fuzzy,
			 int     Weighted, 
			 double   Alpha ) { 
/* ---------------------------------------------------------------------- */
/*                                                                        */
/*  X = pattern to be recognized,					  */
/*  C = class presentative (template),					  */
/*  WeightTable = weights for all vectors in C,				  */
/*  Fuzzy = use fuzzy mapping.						  */
/*  Weighted = use discriminative weights for code vectors                */
/*  Alpha = similarity measure smoothness 				  */
/*									  */
/* ---------------------------------------------------------------------- */

   double  Sim, Weight, tmp_Sum;
   MATRIX x_vec,
	  c_vec;
   int	  i, j, NN;

   if (PDimensionality(X) != PDimensionality(C))
      SegFault("ERROR [MeanWeightedSimilarity]: Dimension mismatch X <-> C");
   else if (MRows(WeightTable) != PNumVectors(C))
      SegFault("ERROR [MeanWeightedSimilarity]: Unequal sizes: weights <-> C");
   else if (MCols(WeightTable) > 1)
      SegFault("ERROR [MeanWeightedSimilarity]: WeightTable must be col. vector");
   else if (Alpha <= 0)
      SegFault("ERROR [MeanWeightedSimilarity]: Alpha should be positive");
   else if (PNumVectors(X) == 0 || PNumVectors(C) == 0)
      SegFault("ERROR [MeanWeightedSimilarity]: X or C is empty.");

   x_vec = NewMatrix(PDimensionality(X),1);
   c_vec = NewMatrix(PDimensionality(C),1);
   Sim = 0;

   NormalizeSumToUnity(WeightTable);

   for(i=1; i <= PNumVectors(X); i++) {
      x_vec = GetVector(x_vec, X, i);
      tmp_Sum = 0;
      if (!Fuzzy) {
         NN = FindNearestVector(x_vec, C);
	 c_vec = GetVector(c_vec, C, NN);
	 if (!Weighted)
	    Weight = 1;
	 else
	    Weight = GetMatrixElement(WeightTable, NN, 1);

	 tmp_Sum += Weight *
	        VectorSimilarity( x_vec, c_vec, Lp_similarity, Alpha );

       }
       else // Fuzzy mapping
	 for(j=1; j <= PNumVectors(C); j++) {
	    if (!Weighted)
	       Weight = 1;
	    else
               Weight = GetMatrixElement(WeightTable, j, 1);
	    tmp_Sum += Weight * 
                   VectorSimilarity( x_vec, c_vec, Lp_similarity, Alpha ); 	
	 }
   
       Sim += tmp_Sum;
   }   
   
   DeleteMatrix(x_vec);
   DeleteMatrix(c_vec);
   
   /* Normalize by vector count */
   return Sim/(1.0 * PNumVectors(X));

}

double MeanWeightedError( PATTERN X, PATTERN C,
                         MATRIX  WeightTable,
                         int     Fuzzy,
			 double	 Alpha ) {
   double  Distortion, 
	  Weight;
   
   MATRIX x_vec,
          c_vec;
   int    i, j, NN;
                
   if (PDimensionality(X) != PDimensionality(C))
      SegFault("ERROR [MeanWeightedError]: Dimension mismatch X <-> C");
   else if (MRows(WeightTable) != PNumVectors(C))
      SegFault("ERROR [MeanWeightedError]: Unequal sizes: weights <-> C");
   else if (MCols(WeightTable) > 1)
      SegFault("ERROR [MeanWeightedError]: WeightTable must be col. vector");
   else if (PNumVectors(X) == 0 || PNumVectors(C) == 0)
      SegFault("ERROR [MeanWeightedError]: X or C is empty.");
          
   x_vec = NewMatrix(PDimensionality(X),1);
   c_vec = NewMatrix(PDimensionality(C),1);
   Distortion = 0;
   
   NormalizeSumToUnity(WeightTable);
   
   for(i=1; i <= PNumVectors(X); i++) {
      x_vec = GetVector(x_vec, X, i);
      if (!Fuzzy) {
         NN = FindNearestVector(x_vec, C);
         c_vec = GetVector(c_vec, C, NN);
         Weight = GetMatrixElement(WeightTable, NN, 1);
	 Weight = exp( -Alpha * Weight );
         Distortion += Weight *
            pow( VectorDistance( x_vec, c_vec, 2), 2);
      }	
      else // Fuzzy mapping
         for(j=1; j <= PNumVectors(C); j++) {
            Weight = GetMatrixElement(WeightTable, j, 1);
	    Weight = exp( -Alpha * Weight );
            Distortion += Weight *
		pow( VectorDistance( x_vec, c_vec, 2), 2);
         }
   }
   
   DeleteMatrix(x_vec);
   DeleteMatrix(c_vec);
      
   /* Normalize by vector count AND weights */  

   if (!Fuzzy)
      return Distortion/(1.0 * PNumVectors(X));
   else {
      Alpha = SumOfElements( WeightTable );
      Weight = exp( -Alpha );
      return Weight*( Distortion/(1.0 * PNumVectors(X)) );
   }
   
}

int NumVectors(PATTERN P)
{
  return P->NumberOfVectors;

}
