/* --------------------------------------------------------------------*/
/*  preprocess.c       Ville Hautamki, <villeh@cs.joensuu.fi>, 2001   */
/*                     Teemu Kilpelinen 15.3.2001                     */
/*                                                                     */
/*  This module includes functions for signal pre-processing.	       */
/*	Version Number:   0.02 (made to work with new framework)       */
/*                        0.01 (original version)		       */
/* --------------------------------------------------------------------*/

#include <stdio.h>
#include <math.h>

#include "signal.h"
#include "speakerid.h"
#include "preprocess.h"


/*	
	Functioning:	Remove the DC-component from signal.  
	Inputs:	        Signal *ptr  Pointer to Signal structure
					
	Outputs:		Signal without DC-component.

	Return value:	-
	History:		Ville Hautamki
	Version:		1.1 (to new framework w/ Signal struct)
                                1.0 (original) (TK 15.09.2000)

	Reference:		
*/
void DCRemove(Signal *ptr)
{
  int i, nsamples;
  double sum = 0;
  double b,c;

  printf("Remove DC-offset...");

  nsamples = get_nsamples(ptr);    	
  
  for (i = 0; i < nsamples; i++)	    
    sum += (get_sample(ptr, i)/nsamples);
  
  for (i=0; i < nsamples; i++) {
    b = get_sample(ptr, i);
    c = b - sum; 
    put_sample(ptr, i, c);
  }
  printf("OK\n");
} 


/*
	Functioning:	Remove lowest frequencies from signal.  
	Inputs:		Pointer to signal struct and high emphasis coeff.
						
	Outputs:		High emphasis filtered signal.

	Return value:	-
	History:		Ville Hautamki
	Version:		1.1 (to new framework w/ Signal struct)
                                1.0 (original) (TK 15.09.2000)
	Reference:		
*/
void HighEmphasisFilter(Signal *ptr, double high_emph_coeff)
{
  int i, nsamples;
  double b,c;
  Signal *sigtmp;

  nsamples = get_nsamples(ptr);	

  /* Memory allocation for tmp-buffer */
  sigtmp = create_signal(nsamples);

  printf("High emphasis filtering...");

  for (i = 0; i < nsamples; i++)
    put_sample(sigtmp, i, get_sample(ptr, i));
        	
  for (i = nsamples - 1; i >= 0; i--) {
    if (i == 0) {                     //Signal boundary condition
      b = high_emph_coeff * get_sample(sigtmp, i);
      c = get_sample(sigtmp, i); 
    } else {
      b = high_emph_coeff * get_sample(sigtmp, i-1);
      c = get_sample(sigtmp, i);
    }
    put_sample(ptr, i, c - b);
  }

  printf("OK\n");
  /* memory deallocation for temp array */
  free_signal(sigtmp);
} 


/*
	Functioning:	Make pre-process process  
	
	Inputs:	  Signal *sigp       Signal datastructure, 
                  Parameters *parap  User parameters
								
	Outputs:		Pre-processed signal and window-function.
	
	Return value:	-
	History:		Teemu Kilpelinen 16.05.2001.
	Version:		1.0 (original)

	Reference:		
*/
void PreProcess(Signal *sigp, Parameters *parap)
{	
  /* Remove DC-offset. Used if wanted. */
  if (parap->featptr->remove_dc_offset) 
    DCRemove(sigp);
  
  /* High emphasis filter. Used if wanted. */
  if (parap->featptr->high_emph_coeff > 0)		
    HighEmphasisFilter(sigp, parap->featptr->high_emph_coeff);	
}
