/* speakerid.c basic API implementation for speaker recognition module. 
   Ville Hautamki, <villeh@cs.joensuu.fi>, 2002                
 */
#include <stdio.h>
#include <stdlib.h>
#include <fftw.h>
#include <float.h>
#include <string.h>

#include "util.h"
#include "signal.h"
#include "speakerid.h"
#include "preprocess.h"
#include "pattern.h"
#include "audio_io.h"
#include "frame.h"
#include "feature_extraction.h"
#include "clustering.h"
#include "distortion.h"
#include "list.h"
#include "csv.h"

#define PATH_BUF 256

char *version = "speakerid: Fri Jun 14 05:21:28 EET DST 2002";

void check_params(Parameters *paraptr);
Element *create_element(Personal *perptr, Parameters *paraptr);
Nbest *insert_nbest(Nbest *listhead, int id, double match, double confidence);
void confidence(Nbest *listhead, double msemin);
PATTERN cb_matching(PATTERN cepstrum, Parameters *paraptr, List *allmodels); 

/* Checks and sets default parameters.
 * Takes as a parameter pointer to Parameters struct
 * Does not return anything.
 */

void check_params(Parameters *paraptr)
{
  char buffer[PATH_BUF];

  /* sanity checks on parameters. If some values are missing then 
     make missing values as defaults. */

  // If model path does not have trailing / insert it.
  if ((paraptr->path_models != NULL) && 
      (paraptr->path_models[strlen(paraptr->path_models)-1] != '/')) {
    strcpy(buffer, paraptr->path_models);
    sprintf(paraptr->path_models, "%s%c", buffer, '/');
  }

  if (paraptr->featptr == NULL) {
    paraptr->featptr = (Featextract *) allocate(sizeof(Featextract),NO_INIT);

  // Global parameters for whole signal
    paraptr->featptr->win_length = 30;
    paraptr->featptr->win_shift_ms = 10;
    paraptr->featptr->remove_sil = 0;
    paraptr->featptr->sil_thr = 1.0;
    paraptr->featptr->remove_dc_offset = 1;
    paraptr->featptr->win_function_type = HAMMING;
    paraptr->featptr->high_emph_coeff = 0.97;

 // Cepstral coeffient parameters
    paraptr->featptr->feat_type = 1;
    paraptr->featptr->num_mel_filters = 0;
    paraptr->featptr->include_delta = 0;
    paraptr->featptr->include_delta_delta = 0;
    paraptr->featptr->include_0th_coeff = 0;
    paraptr->featptr->MFCC_ord = 12;
    paraptr->featptr->lifter_type = 1;
    paraptr->featptr->absolute_magnitude = MAGNITUDESPECTR;
  }

  if (paraptr->modelptr == NULL) {
    paraptr->modelptr = (Modeling *) allocate(sizeof(Modeling),NO_INIT);

    paraptr->modelptr->model_size = 64;
    paraptr->modelptr->max_iterations = 100;
    paraptr->modelptr->stop_threshold = 1e-05;
    paraptr->modelptr->num_initial_solutions = 5; 
  }

  if (paraptr->matchptr == NULL) {
    paraptr->matchptr = (Matching *) allocate(sizeof(Matching),NO_INIT);

    paraptr->matchptr->matching_function = AVDIST;
    paraptr->matchptr->using_cb = 1;
    paraptr->matchptr->maxiter = 5;
    paraptr->matchptr->cb_size = 0;
  }
 
  if (paraptr->deciptr== NULL) {
    paraptr->deciptr = (Decisionlogic *) 
      allocate(sizeof(Decisionlogic),NO_INIT);

    paraptr->deciptr->decision_function = NBEST;
    paraptr->deciptr->num_best = 0;              // 0 == all 
    paraptr->deciptr->confidence_measure = NNRATIO;
  }
}

/* Creates Speaker Database file.
 * Takes as a parameter pointer to Parameters struct.
 * Returns 0 on success and 1 on failure.
 */

int create_speakerdb(Parameters *paraptr)
{
  FILE *dbfile = NULL;

  if (paraptr->speaker_db  == NULL) {
    printf("No Speaker Database file path defined!!");       
    return 1;
  }

  dbfile = open_file(paraptr->speaker_db,"w+");
  fclose(dbfile);

  return 0;
}

/* Creates on speaker database element.
 * Takes as a parameters pointer to Personal struct and pointer to 
 *  Parameters struct.
 * Returns pointer to created Element struct. 
 */

Element *create_element(Personal *perptr, Parameters *paraptr)
{
  Element *elep; 

  elep = (Element *) allocate(sizeof(Element), NO_INIT);
  elep->path = (char *) allocate(sizeof(char)*BUF, NO_INIT);
  elep->name = (char *) allocate(sizeof(char)*BUF, NO_INIT);

  elep->id = perptr->id;
  sprintf(elep->path, "%s%d", paraptr->path_models, perptr->id);
  elep->name = strcpy(elep->name, perptr->name);
  elep->gender = perptr->gender;
  elep->age = perptr->age;
  elep->dialect = perptr->dialect;

  return elep;
}

/* Adds one speaker to speaker database.
 * Takes as a parameter pointer to Parameters struct and pointer to Personal
 *  struct.
 * Returns 0 on success and 1 on failure.
 */

int add_speaker(Parameters *paraptr, Personal *perptr) 
{
  Element *user;
  List *allmodels;
  Signal *sigptr;
  PATTERN cepstrum;
  CLUSTERING model;
  FILE *userfile, *dbfile = NULL;

  printf("Adding new speaker\n");

  check_params(paraptr);

  dbfile = fopen(paraptr->speaker_db,"r");
  allmodels = create_list();
  if (dbfile != NULL) {
    csv_parse(allmodels, dbfile);
    fclose(dbfile);    
  }

  sigptr = ReadSignal(paraptr->instr);
  PreProcess(sigptr, paraptr);
  cepstrum = FeatureExtraction(sigptr, paraptr);

  InitVectorDistortion(vdEuclidean);
  model = NewClustering(paraptr->modelptr->model_size, cepstrum);
  InitClustering(model, cepstrum);
  RGLA(model, cepstrum, paraptr);

  user = create_element(perptr, paraptr);
  insert_element(allmodels, user);

  userfile = open_file(user->path, "w+");
  PrintCodebook(userfile, model);

  dbfile = open_file(paraptr->speaker_db,"w+");
  csv_write(allmodels, dbfile);

  fclose(userfile);
  fclose(dbfile);
  free_list(allmodels);
  DeleteClustering(model);
  DeletePattern(cepstrum);

  return 0;
}

/* Removes one user from speaker database.
 * Takes as a parameter pointer to Parameters struct, and user id
 * Returns 0 on success and 1 on failure.
 */

int remove_speaker(Parameters *paraptr, int id)
{
  List *allmodels;
  Element *user;
  FILE *dbfile;

  check_params(paraptr);
 
  allmodels = create_list();
  dbfile = open_file(paraptr->speaker_db,"r");
  csv_parse(allmodels, dbfile);
  fclose(dbfile);  

  user = find_element(allmodels, id);
  if (user == NULL) {
    printf("remove_speaker: No such speaker!!");
    return(1);
  }

  free_element(allmodels, user);
  dbfile = open_file(paraptr->speaker_db,"w+");
  csv_write(allmodels, dbfile);
 
  fclose(dbfile);
  free_list(allmodels);

  return 0;
}

/* Returns n best list of speakers 
 * Takes as a parameter pointer to Parameters struct.
 * Returns pointer to Nbest lis head. If failure returns NULL.
 */

Nbest *find_most_similar_spkr(Parameters *paraptr)
{
  double mse;
  Iterator *itep;
  Element *user;
  List *allmodels;
  Signal *sigptr;
  PATTERN cepstrum, model, cbcepstrum;
  Nbest *listhead = NULL, *temp;
  FILE *dbfile;

  check_params(paraptr);
  
  allmodels = create_list();
  dbfile = open_file(paraptr->speaker_db,"r");
  csv_parse(allmodels, dbfile);
  fclose(dbfile);  

  sigptr = ReadSignal(paraptr->instr);
  PreProcess(sigptr, paraptr);
  cepstrum = FeatureExtraction(sigptr, paraptr);

  if (paraptr->matchptr->using_cb) 
    cbcepstrum = cb_matching(cepstrum, paraptr, allmodels); 

  itep = iterate_start();
  while ((user = iterate_element(allmodels, itep)) != NULL) {
    model = LoadPattern(user->path);
    if (model == NULL) 
      return NULL;
    if (paraptr->matchptr->using_cb)
      mse = MeanSquaredError(cbcepstrum, model);
    else
      mse = MeanSquaredError(cepstrum, model);
    listhead = insert_nbest(listhead, user->id, mse, 0.0); 

    DeletePattern(model);
  }

  temp = listhead;

  if (cepstrum)
    DeletePattern(cepstrum);
  if (paraptr->matchptr->using_cb) 
    DeletePattern(cbcepstrum);
  free_list(allmodels);

  return listhead;
}

/* Removes Nbest structures from memory.
 * Takas as a parameter pointer to list head.
 * Does not return anything.
 */

void remove_nbest(Nbest *nbestptr)
{
  Nbest *temp;

  while (nbestptr) {
    temp = nbestptr;
    nbestptr = nbestptr->next;
    free(temp);
  }
}

/* Inserts one Nbest struct to Nbest list.
 * Takes as parameters pointer to list head, speaker id, match score and 
 *  confidence.
 * Returns pointer to list head.
 */

Nbest *insert_nbest(Nbest *listhead, int id, double match, double confidence)
{
  Nbest *new;

  new = (Nbest *) allocate(sizeof(Nbest), NO_INIT);

  new->match_score =  match;
  new->confidence = confidence;
  new->id = id;
  new->next = listhead;

  return new;
}


/* Prints out Nbest list.
 * Takes as a parameters pointer to list head.
 * Does not return anything.
 */

void print_nbest(Nbest *listhead)
{  
  while (listhead) {
    printf("%d %f\n", listhead->id, listhead->match_score);
    listhead = listhead->next;
   }
}

/* Return library version string.
 * Does not take any parameters.
 */

char *get_version(void)
{
  return version;
}

/* Creates intermediate codebook for matching.
 * Takes as parameters Cepstrum, pointer to Parameters struct and pointer 
 *  to Speaker database.
 * Returns created model in PATTERN data structure.
 */

PATTERN cb_matching(PATTERN cepstrum, Parameters *paraptr, List *allmodels)
{
  CLUSTERING cbcepstrum;
  PATTERN cbmodel; 
  int cbsize;

  printf("Using Codebook in Matching (instead of cepstrum)\n");
  InitVectorDistortion(vdEuclidean);
  if (paraptr->matchptr->cb_size)
    cbsize = paraptr->matchptr->cb_size;
  else 
    cbsize = PNumVectors(cepstrum)/10;

  cbcepstrum = NewClustering(cbsize, cepstrum);
  paraptr->modelptr->max_iterations = paraptr->matchptr->maxiter; 
  InitClustering(cbcepstrum, cepstrum);
  IterateGLA(cbcepstrum, cepstrum, paraptr);
  cbmodel = Clustering2Pattern(cbcepstrum); 

  return cbmodel;
}

/* Prints features to file.
 * Takes as parameter pointer to Parameters struct.
 * Return 0 on success and 1 on failure.
 */

int print_features(Parameters *paraptr)
{
  Signal *sigptr;
  PATTERN cepstrum;
  FILE *fp;

  fp = fopen(paraptr->featurefile, "w+");
  if (!fp)
    return 1;

  check_params(paraptr);
  sigptr = ReadSignal(paraptr->instr);
  PreProcess(sigptr, paraptr);
  cepstrum = FeatureExtraction(sigptr, paraptr);
  PrintPattern(cepstrum, fp);
  fclose(fp);

  return 0;
}
